# pylint: disable=no-self-use

from unittest.mock import MagicMock

import numpy as np
import pytest
from napari.utils.colormaps import make_colorbar
from qtpy.QtCore import QPoint, Qt
from qtpy.QtGui import QImage

import PartSegData
from PartSeg.common_backend.base_settings import BaseSettings, ColormapDict, ViewSettings
from PartSeg.common_gui.channel_control import ChannelProperty, ColorComboBox, ColorComboBoxGroup
from PartSeg.common_gui.napari_image_view import ImageView
from PartSegCore.color_image.base_colors import starting_colors
from PartSegCore.image_operations import NoiseFilterType
from PartSegImage import TiffImageReader

try:
    from qtpy import PYQT5, PYQT6
except ImportError:  # pragma: no cover
    PYQT5 = True
    PYQT6 = False

if PYQT5 or PYQT6:

    def array_from_image(image: QImage):
        size = image.size().width() * image.size().height()
        return np.frombuffer(image.bits().asstring(size * image.depth() // 8), dtype=np.uint8)

else:

    def array_from_image(image: QImage):
        size = image.size().width() * image.size().height()
        return np.frombuffer(image.bits(), dtype=np.uint8, count=size * image.depth() // 8)


@pytest.fixture()
def base_settings(tmp_path, qapp):
    return BaseSettings(tmp_path)


@pytest.fixture()
def ch_property(base_settings, qtbot):
    ch_prop = ChannelProperty(base_settings, start_name="test")
    qtbot.add_widget(ch_prop)
    return ch_prop


@pytest.fixture()
def image_view(base_settings, ch_property, qtbot):
    image_view = ImageView(base_settings, ch_property, "test")
    qtbot.add_widget(image_view)
    image = TiffImageReader.read_image(PartSegData.segmentation_analysis_default_image)
    with qtbot.waitSignal(image_view.image_added, timeout=10**6):
        base_settings.image = image

    channels_num = image.channels
    assert image_view.channel_control.channels_count == channels_num

    image_view.channel_control.set_active(1)

    return image_view


class TestChannelProperty:
    def test_fail_construct(self, base_settings):
        with pytest.raises(ValueError, match="non empty start_name"):
            ChannelProperty(base_settings, start_name="")

    def test_collapse(self, base_settings, qtbot):
        ch_prop = ChannelProperty(base_settings, start_name="test")
        qtbot.add_widget(ch_prop)
        ch_prop.show()
        assert not ch_prop.collapse_widget.isChecked()
        assert ch_prop.minimum_value.isVisible()
        ch_prop.collapse_widget.setChecked(True)
        assert not ch_prop.minimum_value.isVisible()
        ch_prop.hide()

    def test_get_value_from_settings(self, base_settings, qtbot):
        ch_prop = ChannelProperty(base_settings, start_name="test1")
        base_settings.set_in_profile("test.range_0", (100, 300))
        mock = MagicMock()
        mock.viewer_name = "test"
        ch_prop.register_widget(mock)
        with pytest.raises(ValueError, match="name test already register"):
            ch_prop.register_widget(mock)
        assert ch_prop.minimum_value.value() == 100
        assert ch_prop.maximum_value.value() == 300
        base_settings.set_in_profile("test.range_0", (200, 500))
        assert ch_prop.minimum_value.value() == 200
        assert ch_prop.maximum_value.value() == 500
        base_settings.set_in_profile("test.range_1", (20, 50))
        assert ch_prop.minimum_value.value() == 200
        assert ch_prop.maximum_value.value() == 500
        with pytest.raises(ValueError, match="name test7 not in register"):
            ch_prop.change_current("test7", 1)


class TestColorComboBox:
    def test_base(self, qtbot):
        dkt = ColormapDict({})
        box = ColorComboBox(0, starting_colors, dkt)
        box.show()
        qtbot.add_widget(box)
        with qtbot.waitSignal(box.channel_visible_changed), qtbot.assertNotEmitted(box.clicked):
            qtbot.mouseClick(box.check_box, Qt.LeftButton)
        with qtbot.waitSignal(box.clicked, timeout=1000):
            qtbot.mouseClick(box, Qt.LeftButton, pos=QPoint(5, 5))
        with qtbot.waitSignal(box.clicked):
            qtbot.mouseClick(box, Qt.LeftButton, pos=QPoint(box.width() - 5, 5))
        index = 3
        with qtbot.waitSignal(box.currentTextChanged):
            box.set_color(starting_colors[index])
        img = np.array(make_colorbar(dkt[starting_colors[index]][0], size=(1, 512)))
        assert np.all(array_from_image(box.image) == img.flatten())
        box.hide()

    def test_visibility(self, qtbot):
        dkt = ColormapDict({})
        box = ColorComboBox(0, starting_colors, dkt, lock=True)
        qtbot.add_widget(box)
        box.show()
        qtbot.wait(100)
        assert box.lock.isVisible()
        box.hide()
        box = ColorComboBox(0, starting_colors, dkt, blur=NoiseFilterType.Gauss)
        qtbot.add_widget(box)
        box.show()
        qtbot.wait(100)
        assert box.blur.isVisible()
        box.hide()
        box = ColorComboBox(0, starting_colors, dkt, gamma=2)
        qtbot.add_widget(box)
        box.show()
        qtbot.wait(100)
        assert box.gamma.isVisible()
        box.hide()

    def test_show_frame_arrow(self, qtbot):
        dkt = ColormapDict({})
        box = ColorComboBox(0, starting_colors, dkt)
        qtbot.add_widget(box)
        box.show()
        box.show_arrow = True
        box.repaint()
        qtbot.wait(100)
        box.show_arrow = False
        box.show_frame = True
        box.repaint()
        qtbot.wait(100)
        box.hide()

    def test_change_colors(self, qtbot):
        dkt = ColormapDict({})
        box = ColorComboBox(0, starting_colors, dkt)
        qtbot.add_widget(box)
        box.change_colors(starting_colors[:-1])
        assert box.count() == len(starting_colors) - 1
        box.change_colors(starting_colors[1:])
        assert box.count() == len(starting_colors) - 1

    def test_item_delegate(self, qtbot):
        dkt = ColormapDict({})
        box = ColorComboBox(0, starting_colors, dkt)
        qtbot.add_widget(box)
        box.show()
        box.showPopup()
        qtbot.wait(100)
        box.hide()


class TestColorComboBoxGroup:
    def test_change_channels_num(self, qtbot, image2):
        settings = ViewSettings()
        box = ColorComboBoxGroup(settings, "test", height=30)
        qtbot.add_widget(box)
        box.set_channels(1)
        box.set_channels(4)
        box.set_channels(10)
        box.set_channels(4)
        box.set_channels(10)
        settings.image = image2
        box.update_channels()
        assert box.layout().count() == image2.channels

    def test_update_colormaps(self, qtbot, base_settings):
        box = ColorComboBoxGroup(base_settings, "test", height=30)
        qtbot.add_widget(box)
        box.set_channels(4)
        assert box.current_colormaps == [base_settings.colormap_dict[x][0] for x in starting_colors[:4]]
        box.update_color_list(starting_colors[1:2])
        assert box.current_colors == [starting_colors[1] for _ in range(4)]
        box.update_color_list()
        assert box.layout().itemAt(0).widget().count() == len(starting_colors)

    def test_settings_updated(self, qtbot, base_settings, monkeypatch):
        box = ColorComboBoxGroup(base_settings, "test", height=30)
        box.set_channels(4)
        mock = MagicMock()
        monkeypatch.setattr(box, "parameters_changed", mock)
        base_settings.set_in_profile("test.lock_0", True)
        mock.assert_called_once_with(0)
        dkt = dict(**base_settings.get_from_profile("test"))
        dkt["lock_0"] = False
        dkt["lock_1"] = True
        base_settings.set_in_profile("test", dkt)
        assert mock.call_count == 5
        mock.assert_called_with(3)

    def test_color_combo_box_group(self, qtbot):
        settings = ViewSettings()
        box = ColorComboBoxGroup(settings, "test", height=30)
        qtbot.add_widget(box)
        box.set_channels(3)
        assert len(box.current_colors) == 3
        assert all(isinstance(x, str) for x in box.current_colors)
        with qtbot.waitSignal(box.coloring_update):
            box.layout().itemAt(0).widget().check_box.setChecked(False)
        with qtbot.waitSignal(box.coloring_update):
            box.layout().itemAt(0).widget().setCurrentIndex(2)
        assert box.current_colors[0] is None
        assert all(isinstance(x, str) for x in box.current_colors[1:])

    def test_color_combo_box_group_and_color_preview(self, qtbot):
        settings = ViewSettings()
        ch_property = ChannelProperty(settings, "test")
        box = ColorComboBoxGroup(settings, "test", ch_property, height=30)
        qtbot.add_widget(box)
        qtbot.add_widget(ch_property)
        box.set_channels(3)
        box.set_active(1)
        with qtbot.assert_not_emitted(box.coloring_update), qtbot.assert_not_emitted(box.change_channel):
            ch_property.minimum_value.setValue(10)
            ch_property.minimum_value.setValue(100)

        def check_parameters(name, index):
            return name == "test" and index == 1

        with qtbot.waitSignal(box.coloring_update), qtbot.waitSignal(
            box.change_channel, check_params_cb=check_parameters
        ):
            ch_property.fixed.setChecked(True)

        with qtbot.waitSignal(box.coloring_update), qtbot.waitSignal(
            box.change_channel, check_params_cb=check_parameters
        ):
            ch_property.minimum_value.setValue(10)

        ch_property.maximum_value.setValue(10000)

        with qtbot.waitSignal(box.coloring_update), qtbot.waitSignal(
            box.change_channel, check_params_cb=check_parameters
        ):
            ch_property.maximum_value.setValue(11000)

        with qtbot.waitSignal(box.coloring_update), qtbot.waitSignal(
            box.change_channel, check_params_cb=check_parameters
        ):
            ch_property.fixed.setChecked(False)

        with qtbot.waitSignal(box.coloring_update), qtbot.waitSignal(
            box.change_channel, check_params_cb=check_parameters
        ):
            ch_property.use_filter.setCurrentEnum(NoiseFilterType.Gauss)

        with qtbot.waitSignal(box.coloring_update), qtbot.waitSignal(
            box.change_channel, check_params_cb=check_parameters
        ):
            ch_property.use_filter.setCurrentEnum(NoiseFilterType.Median)

        ch_property.filter_radius.setValue(0.5)
        with qtbot.waitSignal(box.coloring_update), qtbot.waitSignal(
            box.change_channel, check_params_cb=check_parameters
        ):
            ch_property.filter_radius.setValue(2)

        with qtbot.waitSignal(box.coloring_update), qtbot.waitSignal(
            box.change_channel, check_params_cb=check_parameters
        ):
            ch_property.use_filter.setCurrentEnum(NoiseFilterType.No)

        with qtbot.assert_not_emitted(box.coloring_update), qtbot.assert_not_emitted(box.change_channel):
            ch_property.filter_radius.setValue(0.5)

    @pytest.mark.windows_ci_skip()
    @pytest.mark.parametrize("filter_value", NoiseFilterType.__members__.values())
    def test_image_view_integration_filter(self, qtbot, tmp_path, filter_value, ch_property, image_view):
        image_view.channel_control.set_active(1)

        def check_parameters(name, index):
            return name == "test" and index == 1

        if filter_value is NoiseFilterType.No:
            with qtbot.waitSignal(image_view.channel_control.coloring_update), qtbot.waitSignal(
                image_view.channel_control.change_channel, check_params_cb=check_parameters
            ):
                ch_property.use_filter.setCurrentEnum(NoiseFilterType.Gauss)
        with qtbot.waitSignal(image_view.channel_control.coloring_update), qtbot.waitSignal(
            image_view.channel_control.change_channel, check_params_cb=check_parameters
        ):
            ch_property.use_filter.setCurrentEnum(filter_value)
        image4 = image_view.viewer_widget.screenshot()
        assert (filter_value != NoiseFilterType.No and np.any(image4 != 255)) or (
            filter_value == NoiseFilterType.No and np.any(image4 == 255)
        )

    @pytest.mark.windows_ci_skip()
    def test_image_view_integration(self, qtbot, tmp_path, ch_property, image_view):
        image_view.viewer_widget.screenshot(flash=False)
        image1 = image_view.viewer_widget._render()
        assert np.any(image1 != 255)
        ch_property.minimum_value.setValue(100)
        ch_property.maximum_value.setValue(10000)
        ch_property.filter_radius.setValue(0.5)
        image2 = image_view.viewer_widget._render()
        assert np.any(image2 != 255)

        assert np.all(image1 == image2)

        def check_parameters(name, index):
            return name == "test" and index == 1

        # Test fixed range
        with qtbot.waitSignal(image_view.channel_control.coloring_update), qtbot.waitSignal(
            image_view.channel_control.change_channel, check_params_cb=check_parameters
        ):
            ch_property.fixed.setChecked(True)

        image1 = image_view.viewer_widget._render()
        assert np.any(image1 != 255)
        with qtbot.waitSignal(image_view.channel_control.coloring_update), qtbot.waitSignal(
            image_view.channel_control.change_channel, check_params_cb=check_parameters
        ):
            ch_property.minimum_value.setValue(20)
        image2 = image_view.viewer_widget._render()
        assert np.any(image2 != 255)
        assert np.any(image1 != image2)

        with qtbot.waitSignal(image_view.channel_control.coloring_update), qtbot.waitSignal(
            image_view.channel_control.change_channel, check_params_cb=check_parameters
        ):
            ch_property.maximum_value.setValue(11000)
        image3 = image_view.viewer_widget.screenshot(flash=False)
        assert np.any(image3 != 255)
        assert np.any(image2 != image3)
        assert np.any(image1 != image3)

        with qtbot.waitSignal(image_view.channel_control.coloring_update), qtbot.waitSignal(
            image_view.channel_control.change_channel, check_params_cb=check_parameters
        ):
            ch_property.fixed.setChecked(False)

        image1 = image_view.viewer_widget.screenshot(flash=False)
        assert np.any(image1 != 255)
        assert np.any(image1 != image2)
        assert np.any(image1 != image3)
        # Test gauss
        with qtbot.waitSignal(image_view.channel_control.coloring_update), qtbot.waitSignal(
            image_view.channel_control.change_channel, check_params_cb=check_parameters
        ):
            ch_property.use_filter.setCurrentEnum(NoiseFilterType.Gauss)
        image4 = image_view.viewer_widget.screenshot(flash=False)
        assert np.any(image4 != 255)
        assert np.any(image1 != image4)
        assert np.any(image2 != image4)
        assert np.any(image3 != image4)
        with qtbot.waitSignal(image_view.channel_control.coloring_update), qtbot.waitSignal(
            image_view.channel_control.change_channel, check_params_cb=check_parameters
        ):
            ch_property.filter_radius.setValue(1)
        image5 = image_view.viewer_widget.screenshot(flash=False)
        assert np.any(image5 != 255)
        assert np.any(image1 != image5)
        assert np.any(image2 != image5)
        assert np.any(image3 != image5)
        assert np.any(image4 != image5)

    @pytest.mark.windows_ci_skip()
    def test_image_view_integration_gauss(self, qtbot, tmp_path, ch_property, image_view):
        def check_parameters(name, index):
            return name == "test" and index == 1

        ch_property.use_filter.setCurrentEnum(NoiseFilterType.Gauss)
        ch_property.filter_radius.setValue(1)
        # Test gauss and fixed range
        ch_property.minimum_value.setValue(100)
        ch_property.maximum_value.setValue(10000)
        with qtbot.waitSignal(image_view.channel_control.coloring_update), qtbot.waitSignal(
            image_view.channel_control.change_channel, check_params_cb=check_parameters
        ):
            ch_property.fixed.setChecked(True)
        image_view.viewer_widget.screenshot(flash=False)
        image1 = image_view.viewer_widget._render()
        with qtbot.waitSignal(image_view.channel_control.coloring_update), qtbot.waitSignal(
            image_view.channel_control.change_channel, check_params_cb=check_parameters
        ):
            ch_property.minimum_value.setValue(10)
        image2 = image_view.viewer_widget.screenshot()
        assert np.any(image2 != 255)
        assert np.any(image1 != image2)

        with qtbot.waitSignal(image_view.channel_control.coloring_update), qtbot.waitSignal(
            image_view.channel_control.change_channel, check_params_cb=check_parameters
        ):
            ch_property.maximum_value.setValue(11000)
        image3 = image_view.viewer_widget.screenshot()
        assert np.any(image3 != 255)
        assert np.any(image2 != image3)
        assert np.any(image1 != image3)

        with qtbot.waitSignal(image_view.channel_control.coloring_update), qtbot.waitSignal(
            image_view.channel_control.change_channel, check_params_cb=check_parameters
        ):
            ch_property.fixed.setChecked(False)

        image1 = image_view.viewer_widget.screenshot()
        assert np.any(image1 != 255)
        assert np.any(image1 != image2)
        assert np.any(image1 != image3)



import bisect
import json
import typing
from contextlib import suppress
from functools import partial
from io import BytesIO
from math import ceil
from pathlib import Path
from typing import Dict, Iterable, List, Optional, Set, Tuple

import local_migrator
import numpy as np
from fonticon_fa6 import FA6S
from napari.utils import Colormap
from qtpy.QtCore import QPointF, QRect, Qt, Signal
from qtpy.QtGui import (
    QBrush,
    QColor,
    QFont,
    QFontMetrics,
    QHideEvent,
    QMouseEvent,
    QPainter,
    QPaintEvent,
    QResizeEvent,
    QShowEvent,
)
from qtpy.QtWidgets import (
    QCheckBox,
    QColorDialog,
    QGridLayout,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QScrollArea,
    QToolButton,
    QVBoxLayout,
    QWidget,
)
from superqt.fonticon import setTextIcon

from PartSeg.common_backend.base_settings import BaseSettings, ViewSettings
from PartSeg.common_gui.custom_load_dialog import PLoadDialog
from PartSeg.common_gui.custom_save_dialog import PSaveDialog
from PartSeg.common_gui.icon_selector import IconSelector
from PartSeg.common_gui.numpy_qimage import convert_colormap_to_image
from PartSeg.common_gui.qt_util import get_mouse_x, get_mouse_y
from PartSeg.common_gui.universal_gui_part import InfoLabel
from PartSegCore.color_image.base_colors import Color
from PartSegCore.custom_name_generate import custom_name_generate
from PartSegCore.io_utils import IO_LABELS_COLORMAP, LoadBase, SaveBase
from PartSegCore.utils import BaseModel


def color_from_qcolor(color: QColor) -> Color:
    """Convert :py:class:`PyQt5.QtGui.QColor` to :py:class:`.Color`"""
    return Color(red=color.red() / 255, green=color.green() / 255, blue=color.blue() / 255, alpha=color.alpha() / 255)


def qcolor_from_color(color: Color) -> QColor:
    """Convert :py:class:`.Color` to :py:class:`PyQt5.QtGui.QColor`"""
    return QColor(int(color.red * 255), int(color.green * 255), int(color.blue * 255), int(color.alpha * 255))


class ColormapEdit(QWidget):
    """
    Preview of colormap. Double click used for add/remove colors. Single click on marker allows moving them
    """

    double_clicked = Signal(float)  # On double click emit signal with current position factor.

    def __init__(self):
        super().__init__()
        self.color_list: List[Color] = []
        self.position_list: List[float] = []
        self.move_ind = None
        self.image = convert_colormap_to_image(Colormap([(0, 0, 0)]))
        self.setMinimumHeight(60)

    def paintEvent(self, a0: QPaintEvent) -> None:
        painter = QPainter(self)
        margin = 10
        width = self.width() - 2 * margin
        rect = QRect(margin, margin, width, self.height() - 2 * margin)
        painter.drawImage(rect, self.image)
        painter.save()

        for pos_factor in self.position_list:
            pos = width * pos_factor
            point = QPointF(pos + margin, self.height() / 2)
            painter.setBrush(QBrush(Qt.GlobalColor.black))
            painter.drawEllipse(point, 5, 5)
            painter.setBrush(QBrush(Qt.GlobalColor.white))
            painter.drawEllipse(point, 3, 3)

        painter.restore()

    def refresh(self):
        """Recreate presented image and force repaint event"""
        self.image = convert_colormap_to_image(self._colormap())
        self.repaint()

    def _get_color_ind(self, ratio) -> Optional[int]:
        ind = bisect.bisect_left(self.position_list, ratio)
        if len(self.position_list) > ind and abs(self.position_list[ind] - ratio) < 0.01:
            return ind
        if len(self.position_list) > 0 and ind > 0 and abs(self.position_list[ind - 1] - ratio) < 0.01:
            return ind - 1
        return None

    def _get_ratio(self, event: QMouseEvent, margin=10):
        frame_margin = 10
        width = self.width() - 2 * frame_margin
        mouse_x = get_mouse_x(event)
        mouse_y = get_mouse_y(event)
        if mouse_x < margin or mouse_x > self.width() - margin:
            return None
        if mouse_y < margin or mouse_y > self.height() - margin:
            return None
        return (mouse_x - frame_margin) / width

    def mousePressEvent(self, e: QMouseEvent) -> None:
        ratio = self._get_ratio(e, 5)
        if ratio is None:
            return
        ind = self._get_color_ind(ratio)
        if ind is None:
            return
        self.move_ind = ind

    def mouseReleaseEvent(self, e: QMouseEvent) -> None:
        self._move_color(e)
        self.move_ind = None

    def _move_color(self, e: QMouseEvent) -> None:
        ratio = self._get_ratio(e)
        if ratio is None or self.move_ind is None:
            return
        self.position_list.pop(self.move_ind)
        ind = bisect.bisect_left(self.position_list, ratio)
        col = self.color_list.pop(self.move_ind)
        self.color_list.insert(ind, col)
        self.position_list.insert(ind, ratio)
        self.move_ind = ind
        self.refresh()

    def mouseMoveEvent(self, e: QMouseEvent) -> None:
        self._move_color(e)

    def mouseDoubleClickEvent(self, event: QMouseEvent) -> None:
        """
        If click near marker remove it. Otherwise emit `double_click` signal with event position factor.
        """
        ratio = self._get_ratio(event)
        if ratio is None:
            return
        ind = self._get_color_ind(ratio)
        if ind is not None:
            self.position_list.pop(ind)
            self.color_list.pop(ind)
            self.refresh()
            return
        self.double_clicked.emit(ratio)

    def add_color(self, position: float, color: Color):
        """
        Add color to current colormap

        :param color: Color with position.

        """
        ind = bisect.bisect_left(self.position_list, position)
        self.color_list.insert(ind, color)
        self.position_list.insert(ind, position)
        self.refresh()

    def clear(self):
        """
        Remove color markers. Reset to initial state.
        """
        self.color_list = []
        self.position_list = []
        self.image = convert_colormap_to_image(Colormap([(0, 0, 0), (1, 1, 1)]))
        self.repaint()

    def distribute_evenly(self):
        """
        Distribute color markers evenly.
        """
        for i, pos in enumerate(np.linspace(0, 1, len(self.position_list))):
            self.position_list[i] = pos
        self.refresh()

    @property
    def colormap(self) -> Colormap:
        return self._colormap()

    def _colormap(self) -> Colormap:
        """colormap getter"""
        if len(self.color_list) == 0:
            return Colormap("black")

        color_list = [x.as_tuple() for x in self.color_list]
        position_list = self.position_list[:]
        if position_list[0] != 0:
            position_list.insert(0, 0)
            color_list.insert(0, color_list[0])

        if position_list[-1] != 1:
            position_list.append(1)
            color_list.append(color_list[-1])

        return Colormap(colors=color_list, controls=position_list)

    @colormap.setter
    def colormap(self, val: Colormap):
        """colormap setter"""
        self.position_list = val.controls.tolist()
        self.color_list = [Color.from_tuple(x) for x in val.colors.tolist()]
        self.refresh()

    def reverse(self):
        self.position_list = [1 - x for x in reversed(self.position_list)]
        self.color_list = list(reversed(self.color_list))
        self.refresh()


class ColormapCreator(QWidget):
    """
    Widget for creating colormap.
    """

    colormap_selected = Signal(Colormap)
    """
    emitted on save button click. Contains current colormap in format accepted by :py:func:`create_color_map`
    """

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.color_picker = QColorDialog()
        self.color_picker.setWindowFlag(Qt.WindowType.Widget)
        self.color_picker.setOptions(
            QColorDialog.ColorDialogOption.DontUseNativeDialog | QColorDialog.ColorDialogOption.NoButtons
        )
        self.show_colormap = ColormapEdit()
        self.clear_btn = QPushButton("Clear")
        self.save_btn = QPushButton("Save")
        self.distribute_btn = QPushButton("Distribute evenly")
        self.reverse_btn = QPushButton("Reverse")
        self.info_label = InfoLabel(
            [
                "<strong>Tip:</strong> Select color and double click on a color bar below. "
                "Repeat to add another colors.",
                "<strong>Tip:</strong> Double click on a marker to remove it.",
                "<strong>Tip:</strong> Press and hold mouse left button on a marker to drag and drop it on color bar.",
            ],
            delay=10000,
        )
        layout = QVBoxLayout()
        layout.addWidget(self.color_picker)
        layout.addWidget(self.info_label)
        layout.addWidget(self.show_colormap)
        btn_layout = QHBoxLayout()
        btn_layout.addStretch(1)
        btn_layout.addWidget(self.reverse_btn)
        btn_layout.addWidget(self.distribute_btn)
        btn_layout.addWidget(self.clear_btn)
        btn_layout.addWidget(self.save_btn)
        layout.addLayout(btn_layout)
        self._btn_layout = btn_layout
        self.setLayout(layout)
        self.show_colormap.double_clicked.connect(self.add_color)
        self.clear_btn.clicked.connect(self.show_colormap.clear)
        self.save_btn.clicked.connect(self.save)
        self.reverse_btn.clicked.connect(self.show_colormap.reverse)
        self.distribute_btn.clicked.connect(self.show_colormap.distribute_evenly)

    def add_color(self, pos):
        color = self.color_picker.currentColor()
        self.show_colormap.add_color(pos, color_from_qcolor(color))

    def save(self):
        if self.show_colormap.colormap:
            self.colormap_selected.emit(self.show_colormap.colormap)

    def current_colormap(self) -> Colormap:
        """:return: current colormap"""
        return self.show_colormap.colormap

    def set_colormap(self, colormap: Colormap):
        """set current colormap"""
        self.show_colormap.colormap = colormap
        self.show_colormap.refresh()


class PColormapCreatorMid(ColormapCreator):
    """Class to add export and import buttons to ColormapCreator without full PColormapCreator"""

    def __init__(self, settings: BaseSettings):
        super().__init__()
        self.settings = settings
        self.export_btn = QPushButton("Export")
        self.import_btn = QPushButton("Import")

        self._btn_layout.insertWidget(3, self.import_btn)
        self._btn_layout.insertWidget(3, self.export_btn)

        self.export_btn.clicked.connect(self._export_action)
        self.import_btn.clicked.connect(self._import_action)

    def _import_action(self):
        dial = PLoadDialog(ColormapLoad, settings=self.settings, path=IO_LABELS_COLORMAP)
        if dial.exec_():
            res = dial.get_result()
            self.show_colormap.colormap = res.load_class.load(res.load_location)

    def _export_action(self):
        dial = PSaveDialog(
            ColormapSave,
            settings=self.settings,
            path=IO_LABELS_COLORMAP,
        )
        if dial.exec_():
            res = dial.get_result()
            res.save_class.save(res.save_destination, self.show_colormap.colormap, res.parameters)


class PColormapCreator(PColormapCreatorMid):
    """
    :py:class:`~.ColormapCreator` variant which save result in :py:class:`.ViewSettings`
    """

    def __init__(self, settings: BaseSettings):
        super().__init__(settings)
        for i, el in enumerate(settings.get_from_profile("custom_colors", [])):
            self.color_picker.setCustomColor(
                i, qcolor_from_color(el if isinstance(el, Color) else Color.from_tuple(el))
            )
        self.prohibited_names = set(self.settings.colormap_dict.keys())  # Prohibited name is added to reduce
        # probability of colormap cache collision

    def _save_custom_colors(self):
        colors = [color_from_qcolor(self.color_picker.customColor(i)) for i in range(self.color_picker.customCount())]
        self.settings.set_in_profile("custom_colors", colors)

    def hideEvent(self, a0: QHideEvent) -> None:
        """Save custom colors on hide"""
        self._save_custom_colors()

    def save(self):
        if self.show_colormap.colormap:
            rand_name = custom_name_generate(self.prohibited_names, self.settings.colormap_dict)
            save_colormap_in_settings(self.settings, self.show_colormap.colormap, rand_name)
            self.colormap_selected.emit(self.settings.colormap_dict[rand_name][0])


def save_colormap_in_settings(settings: ViewSettings, colormap: Colormap, name: str):
    colors = list(colormap.colors)
    positions = list(colormap.controls)
    if positions[0] != 0:
        positions.insert(0, 0)
        colors.insert(0, colors[0])
    if positions[-1] != 1:
        positions.append(1)
        colors.append(colors[-1])
    settings.colormap_dict[name] = Colormap(colors=np.array(colors), controls=np.array(positions))
    settings.chosen_colormap_change(name, True)


_icon_selector = IconSelector()


class ChannelPreview(QWidget):
    """
    class for preview single colormap. Witch checkbox for change selection.

    :param colormap: colormap to show
    :param accepted: if checkbox should be checked
    :param name: name which will be emitted in all signals as firs argument
    """

    selection_changed = Signal(str, bool)
    """checkbox selection changed (name)"""
    edit_request = Signal([str], [Colormap])
    """send after pressing edit signal (name) (ColorMap object)"""
    remove_request = Signal(str)
    """Signal with name of colormap (name)"""

    def __init__(
        self, colormap: Colormap, accepted: bool, name: str, removable: bool = False, used: bool = False, parent=None
    ):
        super().__init__(parent)
        self.image = convert_colormap_to_image(colormap)
        self.colormap = colormap
        self.name = name
        self.removable = removable
        self.checked = QCheckBox()
        self.checked.setChecked(accepted)
        self.checked.setDisabled(used)
        self.label = QLabel(name)
        self.label.setFixedWidth(150)
        self.setMinimumWidth(80)
        metrics = QFontMetrics(QFont())
        layout = QHBoxLayout()
        layout.addWidget(self.checked)
        layout.addStretch(1)
        self.remove_btn = QToolButton(self)
        setTextIcon(self.remove_btn, FA6S.trash_can, 16)
        if removable:
            self.remove_btn.setToolTip("Remove colormap")
        else:
            self.remove_btn.setToolTip("This colormap is protected")
        self.remove_btn.setEnabled(not accepted and self.removable)

        self.edit_btn = QToolButton(self)
        setTextIcon(self.edit_btn, FA6S.pen, 16)
        layout.addWidget(self.remove_btn)
        layout.addWidget(self.edit_btn)
        layout.addWidget(self.label)
        self.setLayout(layout)
        self.checked.stateChanged.connect(self._selection_changed)
        self.edit_btn.clicked.connect(partial(self.edit_request.emit, name))
        if len(colormap.controls) < 20:
            self.edit_btn.clicked.connect(partial(self.edit_request[Colormap].emit, colormap))
            self.edit_btn.setToolTip("Create colormap base on this")
        else:
            self.edit_btn.setDisabled(True)
            self.edit_btn.setToolTip("This colormap is not editable")
        self.remove_btn.clicked.connect(partial(self.remove_request.emit, name))
        self.setMinimumHeight(max(metrics.height(), self.edit_btn.minimumHeight(), self.checked.minimumHeight()) + 20)
        self.setToolTip(self.name)

    def _selection_changed(self, _=None):
        chk = self.checked.isChecked()
        self.selection_changed.emit(self.name, chk)
        self.remove_btn.setEnabled(not chk and self.removable)

    def set_blocked(self, block):
        """Set if block possibility of remove or uncheck"""
        self.checked.setDisabled(block)
        if self.removable and not block:
            self.remove_btn.setToolTip("Remove colormap")
        else:
            self.remove_btn.setToolTip("This colormap is protected")
            self.remove_btn.setDisabled(True)

    @property
    def state_changed(self):
        """Inner checkbox stateChanged signal"""
        return self.checked.stateChanged

    @property
    def is_checked(self):
        """If colormap is selected"""
        return self.checked.isChecked()

    def set_chosen(self, state: bool):
        """Set selection of check box."""
        self.checked.setChecked(state)

    def paintEvent(self, event: QPaintEvent):
        painter = QPainter(self)
        layout = self.layout()
        first_el = layout.itemAt(0).widget()
        start = 2 * first_el.x() + first_el.width()
        end = self.remove_btn.x() - first_el.x()
        rect = self.rect()
        rect.setX(start)
        rect.setWidth(end - start)
        painter.drawImage(rect, self.image)
        super().paintEvent(event)


class ColormapList(QWidget):
    """
    Show list of colormaps
    """

    edit_signal = Signal(Colormap)
    """Colormap for edit"""

    remove_signal = Signal(str)
    """Name of colormap to remove"""

    visibility_colormap_change = Signal(str, bool)
    """Hide or show colormap"""

    def __init__(
        self, colormap_map: Dict[str, Tuple[Colormap, bool]], selected: Optional[Iterable[str]] = None, parent=None
    ):
        super().__init__(parent=parent)
        self._selected = set() if selected is None else set(selected)
        self._blocked = set()
        self.current_columns = 1
        self.colormap_map = colormap_map
        self._widget_dict: Dict[str, ChannelPreview] = {}
        self.scroll_area = QScrollArea()
        self.central_widget = QWidget(self)
        layout2 = QVBoxLayout()
        self.grid_layout = QGridLayout()
        layout2.addLayout(self.grid_layout)
        layout2.addStretch(1)
        layout2.setContentsMargins(0, 0, 0, 0)

        self.central_widget.setLayout(layout2)
        self.central_widget.setMinimumWidth(300)
        self.scroll_area.setWidget(self.central_widget)
        self.scroll_area.setWidgetResizable(True)

        layout = QVBoxLayout()
        layout.addWidget(self.scroll_area)
        self.setLayout(layout)

    def showEvent(self, event: QShowEvent):
        self.refresh()

    def get_selected(self) -> Set[str]:
        """Already selected colormaps"""
        return self._selected

    def change_selection(self, name, selected):
        if selected:
            self._selected.add(name)
        else:
            self._selected.remove(name)
        self.visibility_colormap_change.emit(name, selected)

    def blocked(self) -> Set[str]:
        """Channels that cannot be turn of and remove"""
        return self._blocked

    def _get_columns(self):
        return max(1, self.width() // 500)

    def resizeEvent(self, event: QResizeEvent):
        if self._get_columns() != self.current_columns:
            self.refresh()
            self.central_widget.repaint()

    def refresh(self):
        layout: QGridLayout = self.grid_layout
        cache_dict: Dict[str, ChannelPreview] = {}
        self._widget_dict = {}
        for _ in range(layout.count()):
            el: ChannelPreview = layout.takeAt(0).widget()
            if el.name in self.colormap_map:
                cache_dict[el.name] = el
            else:
                el.deleteLater()
                el.edit_request[Colormap].disconnect()
                el.remove_request.disconnect()
                el.selection_changed.disconnect()
        selected = self.get_selected()
        blocked = self.blocked()
        columns = self._get_columns()
        for i, (name, (colormap, removable)) in enumerate(self.colormap_map.items()):
            if name in cache_dict:
                widget = cache_dict[name]
                widget.set_blocked(name in blocked)
                widget.set_chosen(name in selected)
            else:
                widget = self._create_colormap_preview(colormap, name, removable)
                widget.edit_request[Colormap].connect(self.edit_signal)
                widget.remove_request.connect(self._remove_request)
                widget.selection_changed.connect(self.change_selection)
            layout.addWidget(widget, i // columns, i % columns)
            self._widget_dict[name] = widget
        widget: QWidget = layout.itemAt(0).widget()
        height = widget.minimumHeight()
        self.current_columns = columns
        self.central_widget.setMinimumHeight((height + 10) * ceil(len(self.colormap_map) / columns))

    def _create_colormap_preview(self, colormap: Colormap, name: str, removable: bool) -> ChannelPreview:
        return ChannelPreview(
            colormap, name in self.get_selected(), name, removable=removable, used=name in self.blocked(), parent=self
        )

    def check_state(self, name: str) -> bool:
        """
        Check state of widget representing given colormap

        :param name: name of colormap which representing widget should be checked
        """
        return self._widget_dict[name].is_checked

    def set_state(self, name: str, state: bool) -> None:
        """
        Set if given colormap is selected

        :param name: name of colormap
        :param state: state to be set
        """
        self._widget_dict[name].set_chosen(state)

    def get_colormap_widget(self, name) -> ChannelPreview:
        """Access to widget showing colormap. Created for testing purpose."""
        return self._widget_dict[name]

    def _remove_request(self, name):
        _, removable = self.colormap_map[name]
        if not removable:
            raise ValueError(f"ColorMap {name} is protected from remove")
        if name not in self.colormap_map:
            raise ValueError(f"color with name {name} not found in any dict")
        del self.colormap_map[name]
        self.refresh()


class PColormapList(ColormapList):
    """
    Show list of colormaps. Integrated with :py:class:`.ViewSettings`

    :param settings: used for store state
    :param control_names: list of names of :py:class:`PartSeg.common_gui.stack_image_view.ImageView`
        for protect used channels from uncheck or remove
    """

    def __init__(self, settings: ViewSettings, control_names: List[str]):
        super().__init__(settings.colormap_dict)
        settings.colormap_dict.colormap_removed.connect(self.refresh)
        settings.colormap_dict.colormap_added.connect(self.refresh)
        settings.colormap_changes.connect(self.refresh)
        self.settings = settings
        self.color_names = control_names

    def get_selected(self) -> Set[str]:
        return set(self.settings.chosen_colormap)

    def change_selection(self, name, selected):
        self.visibility_colormap_change.emit(name, selected)
        self.settings.chosen_colormap_change(name, selected)

    def blocked(self) -> Set[str]:
        # TODO check only currently presented channels
        blocked = set()
        for el in self.color_names:
            num = self.settings.get_from_profile(f"{el}.channels_count", 0)
            for i in range(num):
                blocked.add(self.settings.get_channel_colormap_name(el, i))
        return blocked

    def _change_colormap_visibility(self, name, visible):
        colormaps = set(self.settings.chosen_colormap)
        if visible:
            colormaps.add(name)
        else:
            with suppress(KeyError):
                colormaps.remove(name)
        self.settings.chosen_colormap = sorted(colormaps)


class ColormapSave(SaveBase):
    __argument_class__ = BaseModel

    @classmethod
    def get_short_name(cls):
        return "colormap_json"

    @classmethod
    def save(
        cls,
        save_location: typing.Union[str, BytesIO, Path],
        project_info,
        parameters: Optional[dict] = None,
        range_changed=None,
        step_changed=None,
    ):
        with open(save_location, "w") as f:
            json.dump(project_info, f, cls=local_migrator.Encoder, indent=4)

    @classmethod
    def get_name(cls) -> str:
        return "Colormap json (*.colormap.json)"


class ColormapLoad(LoadBase):
    __argument_class__ = BaseModel

    @classmethod
    def get_short_name(cls):
        return "colormap_json"

    @classmethod
    def load(
        cls,
        load_locations: typing.List[typing.Union[str, BytesIO, Path]],
        range_changed: Optional[typing.Callable[[int, int], typing.Any]] = None,
        step_changed: Optional[typing.Callable[[int], typing.Any]] = None,
        metadata: typing.Optional[dict] = None,
    ) -> Colormap:
        with open(load_locations[0]) as f:
            return json.load(f, object_hook=local_migrator.object_hook)

    @classmethod
    def get_name(cls) -> str:
        return "Colormap json (*.colormap.json)"



import typing
import warnings
from abc import ABC

import numpy as np
import SimpleITK as sitk
from local_migrator import register_class, rename_key, update_argument
from pydantic import Field

from PartSegCore.algorithm_describe_base import AlgorithmDescribeBase, AlgorithmSelection
from PartSegCore.segmentation.algorithm_base import SegmentationLimitException
from PartSegCore.segmentation.utils import close_small_holes
from PartSegCore.utils import BaseModel


class SingleThresholdParams(BaseModel):
    threshold: float = Field(8000.0, ge=-100000, le=100000, title="Threshold", description="Threshold values")


@register_class(version="0.0.1", migrations=[("0.0.1", rename_key("masked", "apply_mask"))])
class SimpleITKThresholdParams128(BaseModel):
    apply_mask: bool = Field(True, description="If apply mask before calculate threshold")
    bins: int = Field(128, title="Histogram bins", ge=8, le=2**16)


@register_class(version="0.0.1", migrations=[("0.0.1", rename_key("masked", "apply_mask"))])
class SimpleITKThresholdParams256(BaseModel):
    apply_mask: bool = Field(True, description="If apply mask before calculate threshold")
    bins: int = Field(128, title="Histogram bins", ge=8, le=2**16)


class MultipleOtsuThresholdParams(BaseModel):
    components: int = Field(2, title="Number of Components", ge=2, lt=100)
    border_component: int = Field(1, title="Border Component", ge=1, lt=100)
    valley: bool = Field(True, title="Valley emphasis")
    bins: int = Field(128, title="Number of histogram bins", ge=8, le=2**16)


class BaseThreshold(AlgorithmDescribeBase, ABC):
    @classmethod
    def calculate_mask(
        cls,
        data: np.ndarray,
        mask: typing.Optional[np.ndarray],
        arguments: BaseModel,
        operator: typing.Callable[[object, object], bool],
    ):
        raise NotImplementedError


class ManualThreshold(BaseThreshold):
    __argument_class__ = SingleThresholdParams

    @classmethod
    def get_name(cls):
        return "Manual"

    @classmethod
    @update_argument("arguments")
    def calculate_mask(
        cls, data: np.ndarray, mask: typing.Optional[np.ndarray], arguments: SingleThresholdParams, operator
    ):
        result = np.array(operator(data, arguments.threshold)).astype(np.uint8)
        if mask is not None:
            result[mask == 0] = 0
        return result, arguments.threshold


class SitkThreshold(BaseThreshold, ABC):
    __argument_class__ = SimpleITKThresholdParams128

    @classmethod
    @update_argument("arguments")
    def calculate_mask(
        cls, data: np.ndarray, mask: typing.Optional[np.ndarray], arguments: SimpleITKThresholdParams128, operator
    ):
        if mask is not None and mask.dtype != np.uint8 and arguments.apply_mask:
            mask = (mask > 0).astype(np.uint8)
        ob, bg, th_op = (0, 1, np.min) if operator(1, 0) else (1, 0, np.max)
        image_sitk = sitk.GetImageFromArray(data)
        if arguments.apply_mask and mask is not None:
            mask_sitk = sitk.GetImageFromArray(mask)
            calculated = cls.calculate_threshold(image_sitk, mask_sitk, ob, bg, arguments.bins, True, 1)
        else:
            calculated = cls.calculate_threshold(image_sitk, ob, bg, arguments.bins)
        result = sitk.GetArrayFromImage(calculated)
        if mask is not None:
            result[mask == 0] = 0
        threshold = th_op(data[result > 0]) if np.any(result) else th_op(-data)
        return result, threshold

    @staticmethod
    def calculate_threshold(*args, **kwargs):
        raise NotImplementedError


class OtsuThreshold(SitkThreshold):
    @classmethod
    def get_name(cls):
        return "Otsu"

    @staticmethod
    def calculate_threshold(*args, **kwargs):
        return sitk.OtsuThreshold(*args)


class LiThreshold(SitkThreshold):
    __argument_class__ = SimpleITKThresholdParams256

    @classmethod
    def get_name(cls):
        return "Li"

    @staticmethod
    def calculate_threshold(*args, **kwargs):
        return sitk.LiThreshold(*args)


class MaximumEntropyThreshold(SitkThreshold):
    __argument_class__ = SimpleITKThresholdParams256

    @classmethod
    def get_name(cls):
        return "Maximum Entropy"

    @staticmethod
    def calculate_threshold(*args, **kwargs):
        return sitk.MaximumEntropyThreshold(*args)


class RenyiEntropyThreshold(SitkThreshold):
    __argument_class__ = SimpleITKThresholdParams256

    @classmethod
    def get_name(cls):
        return "Renyi Entropy"

    @staticmethod
    def calculate_threshold(*args, **kwargs):
        return sitk.RenyiEntropyThreshold(*args)


class ShanbhagThreshold(SitkThreshold):
    __argument_class__ = SimpleITKThresholdParams256

    @classmethod
    def get_name(cls):
        return "Shanbhag"

    @staticmethod
    def calculate_threshold(*args, **kwargs):
        return sitk.ShanbhagThreshold(*args)


class TriangleThreshold(SitkThreshold):
    __argument_class__ = SimpleITKThresholdParams256

    @classmethod
    def get_name(cls):
        return "Triangle"

    @staticmethod
    def calculate_threshold(*args, **kwargs):
        return sitk.TriangleThreshold(*args)


class YenThreshold(SitkThreshold):
    __argument_class__ = SimpleITKThresholdParams256

    @classmethod
    def get_name(cls):
        return "Yen"

    @staticmethod
    def calculate_threshold(*args, **kwargs):
        return sitk.YenThreshold(*args)


class HuangThreshold(SitkThreshold):
    __argument_class__ = SimpleITKThresholdParams256

    @classmethod
    def get_name(cls):
        return "Huang"

    @staticmethod
    def calculate_threshold(*args, **kwargs):
        return sitk.HuangThreshold(*args)


class IntermodesThreshold(SitkThreshold):
    __argument_class__ = SimpleITKThresholdParams256

    @classmethod
    def get_name(cls):
        return "Intermodes"

    @staticmethod
    def calculate_threshold(*args, **kwargs):
        try:
            return sitk.IntermodesThreshold(*args)
        except RuntimeError as e:
            if "Exceeded maximum iterations for histogram smoothing" in e.args[0]:
                raise SegmentationLimitException(*e.args) from e
            raise


class IsoDataThreshold(SitkThreshold):
    __argument_class__ = SimpleITKThresholdParams256

    @classmethod
    def get_name(cls):
        return "Iso Data"

    @staticmethod
    def calculate_threshold(*args, **kwargs):
        return sitk.IsoDataThreshold(*args)


class KittlerIllingworthThreshold(SitkThreshold):
    __argument_class__ = SimpleITKThresholdParams256

    @classmethod
    def get_name(cls):
        return "Kittler Illingworth"

    @staticmethod
    def calculate_threshold(*args, **kwargs):
        try:
            return sitk.KittlerIllingworthThreshold(*args)
        except RuntimeError as e:
            if "sigma2 <= 0" in e.args[0]:
                raise SegmentationLimitException(*e.args) from e
            raise


class MomentsThreshold(SitkThreshold):
    __argument_class__ = SimpleITKThresholdParams256

    @classmethod
    def get_name(cls):
        return "Moments"

    @staticmethod
    def calculate_threshold(*args, **kwargs):
        return sitk.MomentsThreshold(*args)


class MultipleOtsuThreshold(BaseThreshold):
    __argument_class__ = MultipleOtsuThresholdParams

    @classmethod
    def calculate_mask(
        cls,
        data: np.ndarray,
        mask: typing.Optional[np.ndarray],
        arguments: MultipleOtsuThresholdParams,
        operator: typing.Callable[[object, object], bool],
    ):
        cleaned_image_sitk = sitk.GetImageFromArray(data)
        res = sitk.OtsuMultipleThresholds(cleaned_image_sitk, arguments.components, 0, arguments.bins, arguments.valley)
        res = sitk.GetArrayFromImage(res)
        if operator(1, 0):
            res = (res >= arguments.border_component).astype(np.uint8)
            threshold = np.min(data[res > 0]) if np.any(res) else np.max(data)
        else:
            res = (res < arguments.border_component).astype(np.uint8)
            threshold = np.max(data[res > 0]) if np.any(res) else np.min(data)
        if mask is not None:
            res[mask == 0] = 0
        return res, threshold

    @classmethod
    def get_name(cls) -> str:
        return "Multiple Otsu"


class ThresholdSelection(AlgorithmSelection, class_methods=["calculate_mask"], suggested_base_class=BaseThreshold):
    pass


ThresholdSelection.register(ManualThreshold)
ThresholdSelection.register(OtsuThreshold)
ThresholdSelection.register(LiThreshold)
ThresholdSelection.register(RenyiEntropyThreshold)
ThresholdSelection.register(ShanbhagThreshold)
ThresholdSelection.register(TriangleThreshold)
ThresholdSelection.register(YenThreshold)
ThresholdSelection.register(HuangThreshold)
ThresholdSelection.register(IntermodesThreshold)
ThresholdSelection.register(IsoDataThreshold)
ThresholdSelection.register(KittlerIllingworthThreshold)
ThresholdSelection.register(MomentsThreshold)
ThresholdSelection.register(MaximumEntropyThreshold)
ThresholdSelection.register(MultipleOtsuThreshold)


class DoubleThresholdParams(BaseModel):
    core_threshold: ThresholdSelection = ThresholdSelection.get_default()
    base_threshold: ThresholdSelection = ThresholdSelection.get_default()


class DoubleThreshold(BaseThreshold):
    __argument_class__ = DoubleThresholdParams

    @classmethod
    def get_name(cls):
        return "Base/Core"

    @classmethod
    @update_argument("arguments")
    def calculate_mask(
        cls, data: np.ndarray, mask: typing.Optional[np.ndarray], arguments: DoubleThresholdParams, operator
    ):
        thr: BaseThreshold = ThresholdSelection[arguments.core_threshold.name]
        mask1, thr_val1 = thr.calculate_mask(data, mask, arguments.core_threshold.values, operator)

        thr: BaseThreshold = ThresholdSelection[arguments.base_threshold.name]
        mask2, thr_val2 = thr.calculate_mask(data, mask, arguments.base_threshold.values, operator)
        mask2[mask2 > 0] = 1
        mask2[mask1 > 0] = 2
        return mask2, (thr_val1, thr_val2)


class RangeThresholdParams(DoubleThresholdParams):
    core_threshold: ThresholdSelection = Field(default_factory=ThresholdSelection.get_default, title="Upper threshold")
    base_threshold: ThresholdSelection = Field(default_factory=ThresholdSelection.get_default, title="Lower threshold")


class RangeThreshold(DoubleThreshold):
    __argument_class__ = RangeThresholdParams

    @classmethod
    def get_name(cls):
        return "Range"


@register_class(version="0.0.1", migrations=[("0.0.1", rename_key("hist_num", "bins"))])
class DoubleOtsuParams(BaseModel):
    valley: bool = Field(True, title="Valley emphasis")
    bins: int = Field(128, title="Histogram bins", ge=8, le=2**16)


class DoubleOtsu(BaseThreshold):
    __argument_class__ = DoubleOtsuParams

    @classmethod
    def get_name(cls):
        return "Double Otsu"

    @classmethod
    @update_argument("arguments")
    def calculate_mask(
        cls,
        data: np.ndarray,
        mask: typing.Optional[np.ndarray],
        arguments: DoubleOtsuParams,
        operator: typing.Callable[[object, object], bool],
    ):
        cleaned_image_sitk = sitk.GetImageFromArray(data)
        res = sitk.OtsuMultipleThresholds(cleaned_image_sitk, 2, 0, arguments.bins, arguments.valley)
        res = sitk.GetArrayFromImage(res)
        if not np.any(res.flat):
            return res, (0, 0)
        thr1 = data[res == 2].min()
        thr2 = data[res == 1].min()
        return res, (thr1, thr2)


class MultipleOtsuDoubleThresholdParams(BaseModel):
    components: int = Field(2, title="Number of Components", ge=2, lt=100)
    lower_component: int = Field(1, title="Lower Component", ge=1, lt=100)
    upper_component: int = Field(1, title="Upper Component", ge=1, lt=100)
    valley: bool = Field(True, title="Valley emphasis")
    bins: int = Field(128, title="Number of histogram bins", ge=8, le=2**16)


class MultipleOtsu(BaseThreshold):
    __argument_class__ = MultipleOtsuDoubleThresholdParams

    @classmethod
    def get_name(cls):
        return "Multiple Otsu"

    @classmethod
    def calculate_mask(
        cls,
        data: np.ndarray,
        mask: typing.Optional[np.ndarray],
        arguments: MultipleOtsuDoubleThresholdParams,
        operator: typing.Callable[[object, object], bool],
    ):
        cleaned_image_sitk = sitk.GetImageFromArray(data)
        res = sitk.OtsuMultipleThresholds(cleaned_image_sitk, arguments.components, 0, arguments.bins, arguments.valley)
        res = sitk.GetArrayFromImage(res)
        if not np.any(res.flat):
            return res, (0, 0)
        map_component = np.zeros(arguments.components + 1, dtype=np.uint8)
        map_component[: arguments.lower_component] = 0
        map_component[arguments.lower_component : arguments.upper_component] = 1
        map_component[arguments.upper_component :] = 2
        res2 = map_component[res]
        thr1 = data[res2 == 2].min() if np.any(res2 == 2) else data[res2 == 1].max()
        thr2 = data[res2 == 1].min() if np.any(res2 == 1) else data.max()
        return res2, (thr1, thr2)


@register_class(version="0.0.1", migrations=[("0.0.1", rename_key("minimum_radius", "minimum_border_distance"))])
class MaximumDistanceWatershedParams(BaseModel):
    threshold: ThresholdSelection = ThresholdSelection.get_default()
    dilate_radius: int = Field(5, title="Dilate Radius", ge=1, le=100, description="To merge small objects")
    minimum_size: int = Field(100, title="Minimum Size", ge=1, le=1000000, description="To remove small objects")
    minimum_border_distance: int = Field(
        10,
        title="Border Radius",
        ge=0.0,
        le=100.0,
        description="Minimum distance of local maxima from the border. To avoid artifacts",
    )


class MaximumDistanceCore(BaseThreshold):
    """
    This Is algorithm intended to bue used in "* threshold with watershed" algorithms.
    It generates array with three values:

     * 0 - background,
     * 1 - area to watershed,
     * 2 - core objects to start watershed from

    This algorithm is developed to make possible split of almost convex objects that are touching each other.
    Core objects are identified as local maxima of distance from the border.

    To perform this task the following steps are performed:

    1. Thresholding - to detect whole area of objects. This is controlled by ``threshold`` parameter.
    2. Remove small objects - to remove small objects. This is controlled by ``minimum_size` parameter.
    3. Small objects close - to merge small objects. As distance transform is used, it is required small holes.
       This steep closes holes smaller tan 10px.
    4. Distance transform - to find distance from the border
    5. Identify local maxima - to find core objects
    6. Remove local maxima that are too close to the border - to avoid artifacts.
       This distance is controlled by ``minimum_border_distance`` parameter.
    7. Dilate core objects - to make them bigger. For elongated objects it is possible to have multiple local
       maxima along longest axis of object. This step is to merge them.
       This distance is controlled by ``dilate_radius`` parameter.


    This is algorithm that detect core objects
    """

    __argument_class__ = MaximumDistanceWatershedParams

    @classmethod
    def get_name(cls):
        return "Maximum Distance Core"

    @classmethod
    @update_argument("arguments")
    def calculate_mask(
        cls,
        data: np.ndarray,
        mask: typing.Optional[np.ndarray],
        arguments: MaximumDistanceWatershedParams,
        operator: typing.Callable[[object, object], bool],
    ):
        thr: BaseThreshold = ThresholdSelection[arguments.threshold.name]
        mask1, thr_val = thr.calculate_mask(data, mask, arguments.threshold.values, operator)
        mask1 = sitk.GetArrayFromImage(
            sitk.RelabelComponent(sitk.ConnectedComponent(sitk.GetImageFromArray(mask1)), arguments.minimum_size)
        )
        if not np.any(mask1):
            return mask1, (thr_val, thr_val)
        mask2 = close_small_holes((mask1 > 0), 10)
        data = sitk.GetArrayFromImage(sitk.DanielssonDistanceMap(sitk.GetImageFromArray((mask2 == 0).astype(np.uint8))))

        maxima = sitk.GetArrayFromImage(sitk.RegionalMaxima(sitk.GetImageFromArray(data)))
        maxima[data < arguments.minimum_border_distance] = 0

        dilated_maxima = sitk.GetArrayFromImage(
            sitk.BinaryDilate(sitk.GetImageFromArray(maxima), [arguments.dilate_radius] * 3)
        )

        mask1[mask1 > 0] = 1
        mask1[dilated_maxima > 0] = 2
        if operator(0, 1):
            meth = np.amax
        else:
            meth = np.amin
        return mask1, (thr_val, float(meth(data[mask1 == 2]) if np.any(mask1 == 2) else meth(data[mask1 == 1])))


class DoubleThresholdSelection(
    AlgorithmSelection, class_methods=["calculate_mask"], suggested_base_class=BaseThreshold
):
    pass


DoubleThresholdSelection.register(DoubleThreshold)
DoubleThresholdSelection.register(DoubleOtsu)
DoubleThresholdSelection.register(MultipleOtsu)
DoubleThresholdSelection.register(MaximumDistanceCore, old_names=["Maximum Distance Watershed"])


class RangeThresholdSelection(AlgorithmSelection, class_methods=["calculate_mask"], suggested_base_class=BaseThreshold):
    pass


RangeThresholdSelection.register(RangeThreshold)
RangeThresholdSelection.register(DoubleOtsu)


def __getattr__(name):  # pragma: no cover
    if name == "threshold_dict":
        warnings.warn(
            "threshold_dict is deprecated. Please use ThresholdSelection instead", category=FutureWarning, stacklevel=2
        )
        return ThresholdSelection.__register__

    if name == "double_threshold_dict":
        warnings.warn(
            "double_threshold_dict is deprecated. Please use DoubleThresholdSelection instead",
            category=FutureWarning,
            stacklevel=2,
        )
        return DoubleThresholdSelection.__register__

    raise AttributeError(f"module {__name__} has no attribute {name}")



import operator

import numpy as np
import pytest

from PartSegCore.segmentation.algorithm_base import SegmentationLimitException
from PartSegCore.segmentation.threshold import (
    BaseThreshold,
    DoubleThresholdSelection,
    IntermodesThreshold,
    KittlerIllingworthThreshold,
    ThresholdSelection,
)

square = np.zeros((21, 21))
for i, v in [(2, 1000), (4, 10000), (6, 20000), (7, 21000), (8, 22000), (9, 23000)]:
    square[i:-i, i:-i] = v

cube = np.zeros((21, 21, 21))
for i, v in [(2, 1000), (4, 10000), (6, 20000), (7, 21000), (8, 22000), (9, 23000)]:
    cube[i:-i, i:-i, i:-i] = v

empty_square = np.zeros((21, 21))


@pytest.mark.parametrize("method", ThresholdSelection.__register__.values())
@pytest.mark.parametrize("data", [square, cube], ids=["square", "cube"])
@pytest.mark.parametrize("op", [operator.lt, operator.gt])
@pytest.mark.parametrize("masking", [True, False])
def test_threshold(method: BaseThreshold, data, op, masking):
    mask = (data > 0) if masking else None

    try:
        data, thr_info = method.calculate_mask(data=data, mask=mask, arguments=method.get_default_values(), operator=op)
    except SegmentationLimitException:
        if method is KittlerIllingworthThreshold:
            pytest.xfail("KittlerIllingworth sigma problem")
        if method is IntermodesThreshold:
            pytest.xfail("IntermodesThreshold sigma problem")
        raise
    assert isinstance(data, np.ndarray)
    assert isinstance(thr_info, (int, float))


@pytest.mark.parametrize("method", DoubleThresholdSelection.__register__.values())
@pytest.mark.parametrize("data", [square, cube, empty_square], ids=["square", "cube", "empty_square"])
@pytest.mark.parametrize("op", [operator.lt, operator.gt])
@pytest.mark.parametrize("masking", [True, False])
def test_double_threshold(method: BaseThreshold, data, op, masking):
    mask = (data > 0) if masking else None

    data, thr_info = method.calculate_mask(data=data, mask=mask, arguments=method.get_default_values(), operator=op)

    assert isinstance(data, np.ndarray)
    assert isinstance(thr_info[0], (int, float))
    assert isinstance(thr_info[1], (int, float))



import typing
import warnings
from functools import partial

import qtawesome as qta
from napari.utils import Colormap
from qtpy.QtCore import QEvent, QMargins, QModelIndex, QPoint, QPointF, QRect, QSize, Qt, Signal
from qtpy.QtGui import QColor, QIcon, QMouseEvent, QPainter, QPaintEvent, QPen, QPolygonF, QShowEvent
from qtpy.QtWidgets import (
    QCheckBox,
    QComboBox,
    QDoubleSpinBox,
    QGridLayout,
    QHBoxLayout,
    QLabel,
    QListView,
    QStyle,
    QStyledItemDelegate,
    QStyleOptionViewItem,
    QWidget,
)
from superqt import QEnumComboBox

from PartSeg.common_backend.base_settings import ViewSettings
from PartSeg.common_gui.collapse_checkbox import CollapseCheckbox
from PartSeg.common_gui.numpy_qimage import create_colormap_image
from PartSeg.common_gui.qt_util import get_mouse_x
from PartSeg.common_gui.universal_gui_part import CustomSpinBox
from PartSegCore.image_operations import NoiseFilterType

image_dict = {}  # dict to store QImages generated from colormap

ColorMapDict = typing.MutableMapping[str, typing.Tuple[Colormap, bool]]

try:
    from qtpy import PYQT6
except ImportError:  # pragma: no cover
    PYQT6 = False

if PYQT6:

    def _has_focus_state(state: QStyle.StateFlag) -> bool:
        return QStyle.StateFlag.State_HasFocus in state

else:

    def _has_focus_state(state: QStyle.StateFlag) -> bool:
        return bool(state & QStyle.State(QStyle.StateFlag.State_HasFocus))


class ColorStyledDelegate(QStyledItemDelegate):
    """
    Class for paint :py:class:`~.ColorComboBox` elements when list trigger

    :param base_height: height of single list element
    :param color_dict: Dict mapping name to colors
    """

    def __init__(self, base_height: int, color_dict: ColorMapDict, **kwargs):
        super().__init__(**kwargs)
        self.base_height = base_height
        self.color_dict = color_dict

    def paint(self, painter: QPainter, style: QStyleOptionViewItem, model: QModelIndex):
        if model.data() not in image_dict:
            image_dict[model.data()] = create_colormap_image(model.data(), self.color_dict)
        rect = QRect(style.rect.x(), style.rect.y() + 2, style.rect.width() - 150, style.rect.height() - 4)
        painter.drawImage(rect, image_dict[model.data()])
        rect2 = QRect(style.rect.x() + style.rect.width() - 140, style.rect.y() + 2, 140, style.rect.height() - 4)
        painter.drawText(rect2, Qt.AlignmentFlag.AlignLeft, model.data())
        if _has_focus_state(style.state):
            painter.save()
            pen = QPen()
            pen.setWidth(5)
            painter.setPen(pen)
            painter.drawRect(rect)
            painter.restore()

    def sizeHint(self, style: QStyleOptionViewItem, model: QModelIndex):
        res = super().sizeHint(style, model)
        res.setHeight(self.base_height)
        res.setWidth(max(500, res.width()))
        return res


class ColorComboBox(QComboBox):
    """
    Combobox showing colormap instead of text

    :param id_num: id which be emit in signals. Designed to inform which channel information is changed
    :param colors: list of colors which should be able to choose. All needs to be keys in `color_dict`
    :param color_dict: dict from name to colormap definition
    :param colormap: initial colormap
    :param base_height: initial height of widget
    :param lock: show lock padlock to inform that fixed range is used
    :param blur: show info about blur selected
    """

    triangle_width = 20

    clicked = Signal(int)
    """Information about mouse click event on widget"""
    channel_visible_changed = Signal(int, bool)
    """Signal with information about change of channel visibility (ch_num, visible)"""
    channel_colormap_changed = Signal(int, str)
    """Signal with information about colormap change. (ch_num, name_of_colormap)"""

    def __init__(
        self,
        id_num: int,
        colors: typing.List[str],
        color_dict: ColorMapDict,
        colormap: str = "",
        base_height=50,
        lock=False,
        blur=NoiseFilterType.No,
        gamma=1,
    ):
        super().__init__()
        self.id = id_num
        self.check_box = QCheckBox()  # ColorCheckBox(parent=self)
        self.check_box.setChecked(True)
        self.lock = LockedInfoWidget(base_height - 10)
        self.lock.setVisible(lock)
        self.blur = BlurInfoWidget(base_height - 10)
        self.blur.setVisible(blur != NoiseFilterType.No)
        self.gamma = GammaInfoWidget()
        self.gamma.setVisible(gamma != 1)
        self.color_dict = color_dict
        self.colors = colors
        self.addItems(self.colors)
        self.color = colormap or self.itemText(0)
        self.setCurrentText(self.color)
        self.currentTextChanged.connect(self._update_image)
        self.base_height = base_height
        self.show_arrow = False
        self.show_frame = False
        view = QListView()
        view.setMinimumWidth(300)
        view.setItemDelegate(ColorStyledDelegate(self.base_height, color_dict))
        self.setView(view)
        self.image = None  # only for moment, to reduce code repetition

        layout = QHBoxLayout()
        layout.setContentsMargins(7, 0, 0, 0)
        layout.addWidget(self.check_box)
        layout.addWidget(self.lock)
        layout.addWidget(self.blur)
        layout.addWidget(self.gamma)
        layout.addStretch(1)

        self.setLayout(layout)
        self.check_box.stateChanged.connect(partial(self.channel_visible_changed.emit, self.id))
        self.currentTextChanged.connect(partial(self.channel_colormap_changed.emit, self.id))
        self._update_image()

    def change_colors(self, colors: typing.List[str]):
        """change list of colormaps to choose"""
        self.colors = colors
        current_color = self.currentText()
        prev = self.blockSignals(True)
        try:
            index = colors.index(current_color)
        except ValueError:
            index = -1

        self.clear()
        self.addItems(colors)
        if index != -1:
            self.setCurrentIndex(index)
            self.blockSignals(prev)
        else:
            self.blockSignals(prev)
            self._update_image()
            self.repaint()

    def _update_image(self):
        self.color = self.currentText()
        if self.color not in image_dict:
            image_dict[self.color] = create_colormap_image(self.color, self.color_dict)
        self.image = image_dict[self.color]
        self.show_arrow = False

    def enterEvent(self, event: QEvent):
        self.show_arrow = True
        self.repaint()

    def mouseMoveEvent(self, _):
        self.show_arrow = True

    def leaveEvent(self, event: QEvent):
        self.show_arrow = False
        self.repaint()

    def showEvent(self, _event: QShowEvent):
        self.show_arrow = False

    def paintEvent(self, event: QPaintEvent):
        painter = QPainter(self)
        painter.drawImage(self.rect(), self.image)
        if self.show_frame:
            painter.save()
            pen = QPen()
            pen.setWidth(2)
            pen.setColor(QColor("black"))
            painter.setPen(pen)
            rect = QRect(1, 1, self.width() - 2, self.height() - 2)
            painter.drawRect(rect)
            pen.setColor(QColor("white"))
            painter.setPen(pen)
            rect = QRect(3, 3, self.width() - 6, self.height() - 6)
            painter.drawRect(rect)
            painter.restore()
        if self.show_arrow:
            painter.save()
            triangle = QPolygonF()
            dist = 4
            point1 = QPoint(self.width() - self.triangle_width, 0)
            size = QSize(20, self.height() // 2)
            rect = QRect(point1, size)
            painter.fillRect(rect, QColor("white"))
            triangle.append(QPointF(point1) + QPointF(dist, dist))
            triangle.append(QPointF(point1) + QPointF(size.width() - dist, dist))
            triangle.append(QPointF(point1) + QPointF(size.width() // 2, size.height() - dist))
            painter.setBrush(Qt.GlobalColor.black)
            painter.drawPolygon(triangle, Qt.FillRule.WindingFill)
            painter.restore()

    def mousePressEvent(self, event: QMouseEvent):
        if get_mouse_x(event) > self.width() - self.triangle_width:
            super().mousePressEvent(event)
        self.clicked.emit(self.id)

    def minimumSizeHint(self):
        size: QSize = super().minimumSizeHint()
        return QSize(size.width(), max(size.height(), self.base_height))

    def is_checked(self):
        """check if checkbox on widget is checked"""
        return self.check_box.isChecked()

    @property
    def is_lock(self):
        """check if lock property is set"""
        return self.lock.isVisible

    @property
    def is_blur(self):
        """check if blur property is set"""
        return self.blur.isVisible

    @property
    def set_lock(self) -> typing.Callable[[bool], None]:
        """set lock property"""
        return self.lock.setVisible

    @property
    def set_gamma(self) -> typing.Callable[[bool], None]:
        """set lock property"""
        return self.gamma.setVisible

    @property
    def set_blur(self) -> typing.Callable[[bool], None]:
        """set blur property"""
        return self.blur.setVisible

    @property
    def colormap_changed(self):
        """alias for signal, return color name"""
        return self.currentTextChanged

    def set_selection(self, val: bool):
        """set element selected (add frame)"""
        self.show_frame = val
        self.repaint()

    def set_color(self, val: str):
        """set current colormap"""
        self.setCurrentText(val)


class ChannelProperty(QWidget):
    """
    For manipulate channel properties.
        1. Apply gaussian blur to channel
        2. Fixed range for coloring

    In future should be extended

    :param settings: for storing internal state. allow keep state between sessions
    :param start_name: name used to select proper information from settings object.
        Introduced for case with multiple image view.
    """

    def __init__(self, settings: ViewSettings, start_name: str):
        if not start_name:
            raise ValueError("ChannelProperty should have non empty start_name")
        super().__init__()
        self.current_name = start_name
        self.current_channel = 0
        self._settings = settings
        self.widget_dict: typing.Dict[str, ColorComboBoxGroup] = {}

        self.minimum_value = CustomSpinBox(self)
        self.minimum_value.setRange(-(10**6), 10**6)
        self.minimum_value.valueChanged.connect(self.range_changed)
        self.maximum_value = CustomSpinBox(self)
        self.maximum_value.setRange(-(10**6), 10**6)
        self.maximum_value.valueChanged.connect(self.range_changed)
        self.fixed = QCheckBox("Fix range")
        self.fixed.stateChanged.connect(self.lock_channel)
        self.use_filter = QEnumComboBox(enum_class=NoiseFilterType)
        self.use_filter.setToolTip("Only current channel")
        self.filter_radius = QDoubleSpinBox()
        self.filter_radius.setSingleStep(0.1)
        self.filter_radius.valueChanged.connect(self.gauss_radius_changed)
        self.use_filter.currentIndexChanged.connect(self.gauss_use_changed)
        self.gamma_value = QDoubleSpinBox()
        self.gamma_value.setRange(0.01, 100)
        self.gamma_value.setSingleStep(0.1)
        self.gamma_value.valueChanged.connect(self.gamma_value_changed)

        self.collapse_widget = CollapseCheckbox("Channel property")
        self.collapse_widget.add_hide_element(self.minimum_value)
        self.collapse_widget.add_hide_element(self.maximum_value)
        self.collapse_widget.add_hide_element(self.fixed)
        self.collapse_widget.add_hide_element(self.use_filter)
        self.collapse_widget.add_hide_element(self.filter_radius)
        self.collapse_widget.add_hide_element(self.gamma_value)

        self.setup_ui()

    def setup_ui(self):
        layout = QGridLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.collapse_widget, 0, 0, 1, 4)
        label1 = QLabel("Min bright:")
        layout.addWidget(label1, 1, 0)
        layout.addWidget(self.minimum_value, 1, 1)
        label2 = QLabel("Max bright:")
        layout.addWidget(label2, 2, 0)
        layout.addWidget(self.maximum_value, 2, 1)
        layout.addWidget(self.fixed, 1, 2, 1, 2)
        label3 = QLabel("Filter:")
        layout.addWidget(label3, 3, 0, 1, 1)
        layout.addWidget(self.use_filter, 3, 1, 1, 1)
        layout.addWidget(self.filter_radius, 3, 2, 1, 1)
        label4 = QLabel("Gamma:")
        layout.addWidget(label4, 4, 0, 1, 1)
        layout.addWidget(self.gamma_value, 4, 1, 1, 1)
        self.setLayout(layout)

        self.collapse_widget.add_hide_element(label1)
        self.collapse_widget.add_hide_element(label2)
        self.collapse_widget.add_hide_element(label3)
        self.collapse_widget.add_hide_element(label4)

    def send_info(self):
        """send info to"""
        widget = self.widget_dict[self.current_name]
        widget.parameters_changed(self.current_channel)

    def register_widget(self, widget: "ColorComboBoxGroup") -> None:
        """
        Register new viewer by its color combo box group

        :param ColorComboBoxGroup widget: viewer widget for color control
        """
        if widget.viewer_name in self.widget_dict:
            raise ValueError(f"name {widget.viewer_name} already register")
        self.widget_dict[widget.viewer_name] = widget
        self._settings.connect_to_profile(widget.viewer_name, self.refresh_values)
        self.change_current(widget.viewer_name, 0)

    def refresh_values(self, path: typing.Optional[str]):
        if path is None or path.startswith(self.current_name):
            self.change_current(self.current_name, self.current_channel)

    def change_current(self, name: str, channel: int) -> None:
        """
        Change to show values connected with channel `channel` from viewer `viewer`

        :param str name: name of viewer
        :param int channel: channel to which data should be presented
        :rtype: None
        """
        if name not in self.widget_dict:
            raise ValueError(f"name {name} not in register")
        self.current_name = name
        self.current_channel = channel
        block = self.blockSignals(True)
        self.minimum_value.blockSignals(True)
        self.minimum_value.setValue(
            self._settings.get_from_profile(f"{self.current_name}.range_{self.current_channel}", (0, 65000))[0]
        )
        self.minimum_value.blockSignals(False)
        self.maximum_value.setValue(
            self._settings.get_from_profile(f"{self.current_name}.range_{self.current_channel}", (0, 65000))[1]
        )
        self.use_filter.setCurrentEnum(
            self._settings.get_from_profile(
                f"{self.current_name}.use_filter_{self.current_channel}", NoiseFilterType.No
            )
        )
        self.filter_radius.setValue(
            self._settings.get_from_profile(f"{self.current_name}.filter_radius_{self.current_channel}", 1)
        )
        self.fixed.setChecked(
            self._settings.get_from_profile(f"{self.current_name}.lock_{self.current_channel}", False)
        )
        self.gamma_value.setValue(
            self._settings.get_from_profile(f"{self.current_name}.gamma_value_{self.current_channel}", 1)
        )
        self.blockSignals(block)

    def gamma_value_changed(self):
        self._settings.set_in_profile(
            f"{self.current_name}.gamma_value_{self.current_channel}", self.gamma_value.value()
        )
        self.send_info()

    def gauss_radius_changed(self):
        self._settings.set_in_profile(
            f"{self.current_name}.filter_radius_{self.current_channel}", self.filter_radius.value()
        )
        if self.use_filter.currentEnum() != NoiseFilterType.No:
            self.send_info()

    def gauss_use_changed(self):
        self._settings.set_in_profile(
            f"{self.current_name}.use_filter_{self.current_channel}", self.use_filter.currentEnum()
        )
        if self.use_filter.currentEnum() == NoiseFilterType.Median:
            self.filter_radius.setDecimals(0)
            self.filter_radius.setSingleStep(1)
        else:
            self.filter_radius.setDecimals(2)
            self.filter_radius.setSingleStep(0.1)

        self.send_info()

    def lock_channel(self, value):
        self._settings.set_in_profile(f"{self.current_name}.lock_{self.current_channel}", value)
        self.send_info()

    def range_changed(self):
        self._settings.set_in_profile(
            f"{self.current_name}.range_{self.current_channel}",
            (self.minimum_value.value(), self.maximum_value.value()),
        )
        if self.fixed.isChecked():
            self.send_info()


class ColorComboBoxGroup(QWidget):
    """
    Group of :class:`.ColorComboBox` for control visibility and chose colormap for channels.
    """

    coloring_update = Signal()
    """information about global change of coloring"""
    change_channel = Signal(str, int)
    """information which channel change"""

    def __init__(
        self,
        settings: ViewSettings,
        viewer_name: str,
        channel_property: typing.Optional[ChannelProperty] = None,
        height: int = 40,
    ):
        super().__init__()
        self.viewer_name = viewer_name
        self.height = height
        self.settings = settings
        self.settings.colormap_changes.connect(self.update_color_list)
        self.active_box = 0
        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)
        if channel_property is not None:
            channel_property.register_widget(self)
            self.change_channel.connect(channel_property.change_current)
        settings.connect_channel_colormap_name(viewer_name, self.update_colors)
        settings.connect_to_profile(self.viewer_name, self._settings_updated)

    def _settings_updated(self, path: str = ""):
        if "." in path:
            potential_name = path.rsplit(".", maxsplit=1)[-1]
            if "_" not in potential_name:
                return
            name, potential_num = potential_name.rsplit("_", maxsplit=1)
            if potential_num.isnumeric() and name in {"use_filter", "lock", "gamma_value"}:
                self.parameters_changed(int(potential_num))
        else:
            for i in range(self.layout().count()):
                self.parameters_changed(i)

    @property
    def name(self):  # pragma: no cover
        warnings.warn("name is deprecated", DeprecationWarning, stacklevel=2)
        return self.viewer_name

    def update_colors(self):
        """For each channel update colormap to reflect settings"""
        for i in range(self.layout().count()):
            el: ColorComboBox = self.layout().itemAt(i).widget()
            el.setCurrentText(self.settings.get_channel_colormap_name(self.viewer_name, i))

    def update_color_list(self, colors: typing.Optional[typing.List[str]] = None):
        """Update list of available colormaps in each selector"""
        if colors is None:
            colors = self.settings.chosen_colormap
        for i in range(self.layout().count()):
            el: ColorComboBox = self.layout().itemAt(i).widget()
            el.change_colors(colors)

    def update_channels(self):
        """update number of channels base on settings"""
        self.set_channels(self.settings.channels)

    @property
    def channels_count(self):
        return self.layout().count()

    @property
    def selected_colormaps(self) -> typing.List[Colormap]:
        """For each channel give information about selected colormap by name"""
        resp = []
        for i in range(self.layout().count()):
            el: ColorComboBox = self.layout().itemAt(i).widget()
            resp.append(self.settings.colormap_dict[el.currentText()][0])
        return resp

    @property
    def channel_visibility(self) -> typing.List[bool]:
        resp = []
        for i in range(self.layout().count()):
            el: ColorComboBox = self.layout().itemAt(i).widget()
            resp.append(el.is_checked())
        return resp

    @property
    def current_colors(self) -> typing.List[typing.Optional[str]]:
        """List of  current colors. None if channel is not selected."""
        resp = []
        for i in range(self.layout().count()):
            el: ColorComboBox = self.layout().itemAt(i).widget()
            if el.is_checked():
                resp.append(el.currentText())
            else:
                resp.append(None)
        return resp

    @property
    def current_colormaps(self) -> typing.List[typing.Optional[Colormap]]:
        """List of current colormaps. None if channel is not selected"""
        resp = []
        for i in range(self.layout().count()):
            el: ColorComboBox = self.layout().itemAt(i).widget()
            if el.is_checked():
                resp.append(self.settings.colormap_dict[el.currentText()][0])
            else:
                resp.append(None)
        return resp

    def visibility_changed(self, index):
        self.coloring_update.emit()
        self.change_channel.emit(self.viewer_name, index)

    def change_selected_color(self, index, color):
        self.settings.set_channel_colormap_name(self.viewer_name, index, str(color))
        self.coloring_update.emit()
        self.change_channel.emit(self.viewer_name, index)

    def active_channel(self, channel: int):
        return self.layout().itemAt(channel).widget().is_checked()

    def set_channels(self, num: int):
        """Set number of channels to display"""
        self.settings.set_in_profile(f"{self.viewer_name}.channels_count", num)
        if num >= self.layout().count():
            for i in range(self.layout().count(), num):
                el = ColorComboBox(
                    id_num=i,
                    colors=self.settings.chosen_colormap,
                    color_dict=self.settings.colormap_dict,
                    colormap=self.settings.get_channel_colormap_name(self.viewer_name, i),
                    base_height=self.height,
                    lock=self.settings.get_from_profile(f"{self.viewer_name}.lock_{i}", False),
                    blur=self.settings.get_from_profile(f"{self.viewer_name}.use_filter_{i}", NoiseFilterType.No),
                    gamma=self.settings.get_from_profile(f"{self.viewer_name}.gamma_value_{i}", 1),
                )
                el.clicked.connect(self.set_active)
                el.channel_visible_changed.connect(self.visibility_changed)
                el.channel_colormap_changed.connect(self.change_selected_color)
                self.layout().addWidget(el)
        else:
            for _ in range(self.layout().count() - num):
                el = self.layout().takeAt(num).widget()
                el.colormap_changed.disconnect()
                el.channel_colormap_changed.disconnect()
                el.clicked.disconnect()
                el.channel_visible_changed.disconnect()
                el.deleteLater()
        if num <= self.active_box:
            self.set_active(num - 1)
        self.change_channel.emit(self.viewer_name, self.active_box)

    def set_active(self, pos: int):
        self.active_box = pos
        for i in range(self.layout().count()):
            el = self.layout().itemAt(i).widget()
            el.show_frame = i == self.active_box
        self.change_channel.emit(self.viewer_name, pos)
        self.repaint()

    def get_filter(self) -> typing.List[typing.Tuple[NoiseFilterType, float]]:
        return [
            (
                self.settings.get_from_profile(f"{self.viewer_name}.use_filter_{i}", NoiseFilterType.No),
                self.settings.get_from_profile(f"{self.viewer_name}.filter_radius_{i}", 1),
            )
            for i in range(self.layout().count())
        ]

    def get_limits(self) -> typing.List[typing.Union[typing.Tuple[int, int], None]]:
        resp: typing.List[typing.Union[typing.Tuple[int, int], None]] = [(0, 0)] * self.layout().count()
        for i in range(self.layout().count()):
            resp[i] = (
                self.settings.get_from_profile(f"{self.viewer_name}.range_{i}", (0, 65000))
                if self.settings.get_from_profile(f"{self.viewer_name}.lock_{i}", False)
                else None
            )

        return resp

    def get_gamma(self) -> typing.List[float]:
        return [
            self.settings.get_from_profile(f"{self.viewer_name}.gamma_value_{i}", 1)
            for i in range(self.layout().count())
        ]

    def parameters_changed(self, channel):
        """for ChannelProperty to inform about change of parameters"""
        if self.layout().itemAt(channel) is None:
            return
        widget: ColorComboBox = self.layout().itemAt(channel).widget()
        widget.set_blur(
            self.settings.get_from_profile(f"{self.viewer_name}.use_filter_{channel}", NoiseFilterType.No)
            != NoiseFilterType.No
        )
        widget.set_lock(self.settings.get_from_profile(f"{self.viewer_name}.lock_{channel}", False))
        widget.set_gamma(self.settings.get_from_profile(f"{self.viewer_name}.gamma_value_{channel}", 1) != 1)
        if self.active_channel(channel):
            self.coloring_update.emit()
        if self.active_box == channel:
            self.change_channel.emit(self.viewer_name, channel)


class LockedInfoWidget(QWidget):
    """
    Widget used to present info about lock selection in class :py:class:`~.ColorComboBox`.
    """

    def __init__(self, size=25, margin=1):
        super().__init__()
        self.margin = QMargins(margin, margin, margin, margin)
        self.setFixedWidth(size)
        self.setFixedHeight(size)
        self.setToolTip("Fixed range")

    def paintEvent(self, a0: QPaintEvent) -> None:
        super().paintEvent(a0)
        painter = QPainter(self)
        painter.save()
        painter.setPen(QColor("white"))
        painter.setBrush(QColor("white"))
        painter.drawRect(self.rect() - self.margin)
        icon: QIcon = qta.icon("fa5s.lock", color="black")
        icon.paint(painter, self.rect() - self.margin * 2)
        painter.restore()


class BlurInfoWidget(QWidget):
    """
    Widget used to present info about blur selection in class :py:class:`~.ColorComboBox`.
    """

    def __init__(self, size=25, margin=1):
        super().__init__()
        self.margin = QMargins(margin, margin, margin, margin)
        self.setFixedWidth(size)
        self.setFixedHeight(size)
        self.setToolTip("Filtered")

    def paintEvent(self, a0: QPaintEvent) -> None:
        super().paintEvent(a0)
        painter = QPainter(self)
        painter.save()
        painter.setPen(QColor("white"))
        painter.setBrush(QColor("white"))
        painter.drawRect(self.rect() - self.margin)
        icon: QIcon = qta.icon("mdi.blur", color="black")
        icon.paint(painter, self.rect())
        painter.restore()


class GammaInfoWidget(QWidget):
    def __init__(self, size=25, margin=2):
        super().__init__()
        self.margin = QMargins(margin, margin, margin, margin)
        self.setFixedWidth(size)
        self.setFixedHeight(size)
        self.setToolTip("Gamma translated")

    def paintEvent(self, a0: QPaintEvent) -> None:
        super().paintEvent(a0)
        painter = QPainter(self)
        painter.save()
        painter.setPen(QColor("white"))
        painter.setBrush(QColor("white"))
        painter.drawRect(self.rect() - self.margin)
        icon: QIcon = qta.icon("mdi.chart-bell-curve-cumulative", color="black")
        icon.paint(painter, self.rect())
        painter.restore()



import packaging.version
import pytest

import PartSeg


def test_analysis_import():
    try:
        import PartSeg._roi_analysis.main_window  # noqa: F401
    except ImportError:  # pragma: no cover
        pytest.fail("Error in importing segmentation ui")


def test_launcher_import():
    try:
        import PartSeg._launcher.main_window  # noqa: F401
    except ImportError:  # pragma: no cover
        pytest.fail("Error in importing launcher ui")


def test_segmentation_import():
    try:
        import PartSeg._roi_mask.main_window  # noqa: F401
    except ImportError:  # pragma: no cover
        pytest.fail("Error in importing mask segmentation ui")


def test_version_string():
    assert isinstance(packaging.version.parse(PartSeg.__version__), packaging.version.Version)



import pytest
from pydantic import BaseModel
from qtpy import API

from PartSeg._launcher.main_window import MainWindow as LauncherMain
from PartSeg._roi_analysis.main_window import MainWindow as AnalysisMain
from PartSeg._roi_mask.main_window import MainWindow as MaskMain
from PartSeg.common_backend.base_argparser import CustomParser
from PartSeg.launcher_main import create_parser, select_window


class ArgsMock(BaseModel):
    gui: str
    mf: bool = False
    image: str = ""
    batch: bool = False


def test_create_parser():
    assert isinstance(create_parser(), CustomParser)


@pytest.mark.skipif(API.startswith("pyside"), reason="problem with make_napari_viewer fixture")
@pytest.mark.parametrize(
    ("gui", "klass"), [("roi_analysis", AnalysisMain), ("roi_mask", MaskMain), ("launcher", LauncherMain)]
)
def test_select_window(qtbot, gui, klass):
    args = ArgsMock(gui=gui)
    widget = select_window(args)
    assert isinstance(widget, klass)
    widget.close()
    widget.deleteLater()
    qtbot.wait(50)



# pylint: disable=no-self-use

import numpy as np
import pytest

from PartSegCore.image_operations import gaussian, median


class TestImageOperation:
    @pytest.mark.parametrize("dims", [2, 3, 4])
    @pytest.mark.parametrize("method", [gaussian, median])
    @pytest.mark.parametrize("per_layer", [True, False])
    def test_filter_dims_dummy(self, dims, method, per_layer):
        data = np.zeros((2,) * (dims - 2) + (5, 5), dtype=np.uint8)
        res = method(data, 1, per_layer)
        assert np.all(res == 0)

    @pytest.mark.parametrize("dims", [2, 3, 4])
    @pytest.mark.parametrize("method", [gaussian, median])
    @pytest.mark.parametrize("per_layer", [True, False])
    def test_filter_dims_ones(self, dims, method, per_layer):
        data = np.zeros((2,) * (dims - 2) + (5, 5), dtype=np.uint8)
        slices = (0,) * (dims - 2) + (1,)
        data[slices] = 1
        res = method(data, 2, per_layer)
        assert not np.all(res == data)



from PartSegCore.color_image.base_colors import default_colormap_dict, default_label_dict
from PartSegCore_compiled_backend.color_image_cython import add_labels, calculate_borders, color_grayscale, resolution

__all__ = (
    "add_labels",
    "calculate_borders",
    "color_grayscale",
    "default_colormap_dict",
    "default_label_dict",
    "resolution",
)



import os
from typing import Any, Optional

import numpy as np

from PartSegCore.mask.io_functions import MaskProjectTuple, SaveROI, SaveROIOptions
from PartSegCore.roi_info import ROIInfo
from PartSegImage.image import DEFAULT_SCALE_FACTOR


def napari_write_labels(path: str, data: Any, meta: dict) -> Optional[str]:
    if not isinstance(data, np.ndarray):
        return None
    ext = os.path.splitext(path)[1]
    if ext in SaveROI.get_extensions():
        project = MaskProjectTuple(file_path="", image=None, roi_info=ROIInfo(data))
        SaveROI.save(
            path, project, parameters=SaveROIOptions(spacing=list(np.divide(meta["scale"], DEFAULT_SCALE_FACTOR)[-3:]))
        )
        return path
    return None



import collections.abc
import inspect
import logging
import typing
from contextlib import suppress
from copy import deepcopy
from enum import Enum

import numpy as np
from magicgui.widgets import ComboBox, EmptyWidget, Widget, create_widget
from napari.layers.base import Layer
from pydantic import BaseModel
from qtpy.QtCore import QMargins, QObject, Signal
from qtpy.QtGui import QHideEvent, QPainter, QPaintEvent, QResizeEvent
from qtpy.QtWidgets import (
    QApplication,
    QCheckBox,
    QComboBox,
    QDoubleSpinBox,
    QFormLayout,
    QLabel,
    QLineEdit,
    QMessageBox,
    QScrollArea,
    QSizePolicy,
    QSpinBox,
    QStackedLayout,
    QVBoxLayout,
    QWidget,
)
from sentry_sdk.integrations.logging import ignore_logger
from superqt import QEnumComboBox

from PartSeg.common_backend.base_settings import BaseSettings
from PartSeg.common_backend.segmentation_thread import SegmentationThread
from PartSeg.common_gui.error_report import ErrorDialog
from PartSeg.common_gui.universal_gui_part import ChannelComboBox, CustomDoubleSpinBox, CustomSpinBox, Hline
from PartSegCore.algorithm_describe_base import (
    AlgorithmDescribeBase,
    AlgorithmProperty,
    AlgorithmSelection,
    ROIExtractionProfile,
    base_model_to_algorithm_property,
)
from PartSegCore.segmentation.algorithm_base import (
    ROIExtractionAlgorithm,
    ROIExtractionResult,
    SegmentationLimitException,
)
from PartSegImage import Channel, Image

try:
    from pydantic.fields import UndefinedType
except ImportError:  # pragma: no cover
    from pydantic_core import PydanticUndefinedType as UndefinedType

logger = logging.getLogger(__name__)
ignore_logger(__name__)


def recursive_update(d, u):
    if not isinstance(d, typing.MutableMapping):
        d = {}
    for k, v in u.items():
        d[k] = recursive_update(d.get(k, {}), v) if isinstance(v, collections.abc.Mapping) else v
    return d


def _pretty_print(data, indent=2) -> str:
    if isinstance(data, typing.Mapping):
        res = "\n"
        for k, v in data.items():
            res += f"{' ' * indent}{k}: {_pretty_print(v, indent+2)}\n"
        return res[:-1]
    return str(data)


class ProfileSelect(QComboBox):
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self._settings = None

    def _update_choices(self):
        if hasattr(self._settings, "roi_profiles"):
            self.clear()
            self.addItems(list(self._settings.roi_profiles.keys()))

    def set_settings(self, settings: BaseSettings):
        self._settings = settings
        if hasattr(self._settings, "roi_profiles"):
            self._settings.roi_profiles.setted.connect(self._update_choices)
            self._settings.roi_profiles.deleted.connect(self._update_choices)
        self._update_choices()

    def get_value(self):
        if self._settings is not None and hasattr(self._settings, "roi_profiles") and self.currentText():
            return self._settings.roi_profiles[self.currentText()]
        return None

    def set_value(self, value):
        if (
            self._settings is not None
            and hasattr(self._settings, "roi_profiles")
            and value.name in self._settings.roi_profiles
        ):
            self.setCurrentText(value.name)


class QtAlgorithmProperty(AlgorithmProperty):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._widget = self._get_field()
        self.change_fun = self.get_change_signal(self._widget)
        self._getter, self._setter = self.get_getter_and_setter_function(self._widget)
        with suppress(TypeError, ValueError):
            self._setter(self._widget, self.default_value)

    def get_value(self):
        return self._getter(self._widget)

    def is_multiline(self):
        return getattr(self._widget, "__multiline__", False)

    def recursive_get_values(self):
        if isinstance(self._widget, SubAlgorithmWidget):
            return self._widget.recursive_get_values()
        return self.get_value()

    def set_value(self, val):
        """set value of widget"""
        try:
            return self._setter(self._widget, val)
        except (TypeError, ValueError) as e:
            logger.error("Error %s setting value %s to %s", e, val, self.name)

    def get_field(self) -> QWidget:
        """
        Get representing widget
        :return:
        :rtype:
        """
        return self._widget

    @classmethod
    def from_algorithm_property(cls, ob: typing.Union[str, AlgorithmProperty]):
        """
        Create class instance base on :py:class:`.AlgorithmProperty` instance
        If ob is string equal to `hline` or that contains only
        `-` of length at least 5 then return :py:class:`.HLine`

        :type ob: AlgorithmProperty | str
        :param ob: AlgorithmProperty object or label
        :return: QtAlgorithmProperty | QLabel
        """
        if isinstance(ob, AlgorithmProperty):
            return cls(
                name=ob.name,
                user_name=ob.user_name,
                default_value=ob.default_value,
                options_range=ob.range,
                value_type=ob.value_type,
                possible_values=ob.possible_values,
                help_text=ob.help_text,
                per_dimension=ob.per_dimension,
                mgi_options=ob.mgi_options,
            )
        if isinstance(ob, str):
            if ob.lower() == "hline" or len(ob) > 5 and all(x == "-" for x in ob):
                return Hline()
            return QLabel(ob)
        raise ValueError(f"unknown parameter type {type(ob)} of {ob}")

    @staticmethod
    def _get_numeric_field(ap: AlgorithmProperty):
        if issubclass(ap.value_type, int):
            res = CustomSpinBox()
            if not isinstance(ap.default_value, int):
                raise ValueError(
                    f"Incompatible types. default_value should be type of int. Is {type(ap.default_value)}"
                )
        else:  # issubclass(ap.value_type, float):
            res = CustomDoubleSpinBox()
            if not isinstance(ap.default_value, (int, float)):
                raise ValueError(
                    f"Incompatible types. default_value should be type of float. Is {type(ap.default_value)}"
                )
        if ap.default_value < res.minimum():
            res.setMinimum(ap.default_value)
        if ap.range is not None:
            res.setRange(*ap.range)
        return res

    @classmethod
    def _get_field_from_value_type(cls, ap: AlgorithmProperty):
        if issubclass(ap.value_type, Channel):
            res = ChannelComboBox(parent=None)
            res.change_channels_num(10)
        elif issubclass(ap.value_type, AlgorithmDescribeBase):
            res = SubAlgorithmWidget(ap)
        elif issubclass(ap.value_type, bool):
            res = QCheckBox()
        elif issubclass(ap.value_type, (float, int)):
            res = cls._get_numeric_field(ap)
        elif issubclass(ap.value_type, Enum):
            # noinspection PyTypeChecker
            res = QEnumComboBox(enum_class=ap.value_type)
        elif issubclass(ap.value_type, str):
            res = QLineEdit()
        elif issubclass(ap.value_type, ROIExtractionProfile):
            res = ProfileSelect()
        elif issubclass(ap.value_type, list):
            res = QComboBox(parent=None)
            res.addItems(list(map(str, ap.possible_values)))
        elif issubclass(ap.value_type, BaseModel):
            res = FieldsList([cls.from_algorithm_property(x) for x in base_model_to_algorithm_property(ap.value_type)])
        else:
            res = cls._get_field_magicgui(ap)
        return res

    @classmethod
    def _get_field_magicgui(cls, ap: AlgorithmProperty) -> Widget:
        if isinstance(ap.default_value, UndefinedType) or ap.default_value is Ellipsis:
            res = create_widget(annotation=ap.value_type, options=ap.mgi_options)
        else:
            try:
                res = create_widget(value=ap.default_value, annotation=ap.value_type, options=ap.mgi_options)
            except ValueError as e:
                if "None is not a valid choice." in str(e):
                    res = create_widget(annotation=ap.value_type, options=ap.mgi_options)
                else:  # pragma: no cover
                    raise e

        if isinstance(res, EmptyWidget):  # pragma: no cover
            raise ValueError(f"Unknown type {ap.value_type}")
        return res

    def _get_field(self) -> typing.Union[QWidget, Widget]:
        """
        Get proper widget for given field type. Overwrite if would like to support new data types.
        """
        if self.per_dimension:
            self.per_dimension = False
            prop = self.from_algorithm_property(self)
            self.per_dimension = True
            res = ListInput(prop, 3)
        elif not inspect.isclass(self.value_type):
            res = self._get_field_magicgui(self)
        elif hasattr(self.value_type, "get_object"):
            res = self.value_type.get_object()
        else:
            res = self._get_field_from_value_type(self)
        tool_tip_text = self.help_text or ""
        tool_tip_text += f" default value: {_pretty_print(self.default_value)}"
        if isinstance(res, QWidget):
            res.setToolTip(tool_tip_text)
        if isinstance(res, Widget):
            res.tooltip = tool_tip_text  # pylint: disable=attribute-defined-outside-init # false positive
        return res

    @staticmethod
    def get_change_signal(widget: typing.Union[QWidget, Widget]):  # noqa: PLR0911
        if isinstance(widget, Widget):
            return widget.changed
        if isinstance(widget, QComboBox):
            return widget.currentIndexChanged
        if isinstance(widget, QCheckBox):
            return widget.stateChanged
        if isinstance(widget, (QSpinBox, QDoubleSpinBox)):
            return widget.valueChanged
        if isinstance(widget, QLineEdit):
            return widget.textChanged
        if isinstance(widget, SubAlgorithmWidget):
            return widget.values_changed
        if isinstance(widget, ListInput):
            return widget.change_signal
        if isinstance(widget, FieldsList):
            return widget.changed
        if hasattr(widget, "values_changed"):
            return widget.values_changed
        raise ValueError(f"Unsupported type: {type(widget)}")

    @staticmethod
    def get_getter_and_setter_function(  # noqa: PLR0911
        widget: typing.Union[QWidget, Widget],
    ) -> typing.Tuple[
        typing.Callable[
            [typing.Union[QWidget, Widget]],
            typing.Any,
        ],
        typing.Callable[[typing.Union[QWidget, Widget], typing.Any], None],
    ]:
        """
        For each widget type return proper functions. This functions need instance as first argument

        :return: (getter, setter)
        """
        if isinstance(widget, Widget):
            return _value_get, _value_set
        if isinstance(widget, (ProfileSelect, ChannelComboBox)):
            return widget.__class__.get_value, widget.__class__.set_value
        if isinstance(widget, QEnumComboBox):
            return widget.__class__.currentEnum, widget.__class__.setCurrentEnum
        if isinstance(widget, QComboBox):
            return widget.__class__.currentText, widget.__class__.setCurrentText
        if isinstance(widget, QCheckBox):
            return widget.__class__.isChecked, widget.__class__.setChecked
        if isinstance(widget, (QSpinBox, QDoubleSpinBox)):
            return widget.__class__.value, widget.__class__.setValue
        if isinstance(widget, QLineEdit):
            return widget.__class__.text, widget.__class__.setText
        if isinstance(widget, SubAlgorithmWidget):
            return widget.__class__.get_values, widget.__class__.set_values
        if isinstance(widget, ListInput):
            return widget.__class__.get_value, widget.__class__.set_value
        if hasattr(widget.__class__, "get_value") and hasattr(widget.__class__, "set_value"):
            return widget.__class__.get_value, widget.__class__.set_value
        raise ValueError(f"Unsupported type: {type(widget)}")


class FieldsList(QObject):
    changed = Signal()

    def __init__(self, field_list: typing.List[QtAlgorithmProperty]):
        super().__init__()
        self.field_list = field_list
        for el in field_list:
            el.change_fun.connect(self._changed_wrap)

    def get_value(self):
        return {el.name: el.get_value() for el in self.field_list}

    def _changed_wrap(self, val=None):
        self.changed.emit()

    def set_value(self, val):
        if isinstance(val, dict):
            self._set_value_dkt(val)
        else:
            self._set_value_base_model(val)

    def _set_value_base_model(self, val):
        for el in self.field_list:
            if hasattr(val, el.name):
                el.set_value(getattr(val, el.name))

    def _set_value_dkt(self, val: dict):
        for el in self.field_list:
            if el.name in val:
                el.set_value(val[el.name])


class ListInput(QWidget):
    change_signal = Signal()

    def __init__(self, property_el: QtAlgorithmProperty, length):
        super().__init__()
        self.input_list = [property_el.from_algorithm_property(property_el) for _ in range(length)]
        layout = QVBoxLayout()
        for el in self.input_list:
            el.change_fun.connect(_any_arguments(self.change_signal.emit))
            layout.addWidget(el.get_field())
        self.setLayout(layout)

    def get_value(self):
        return [x.get_value() for x in self.input_list]

    def set_value(self, value):
        if not isinstance(value, (list, tuple)):
            value = [value for _ in range(len(self.input_list))]
        for f, val in zip(self.input_list, value):
            f.set_value(val)


def _any_arguments(fun):
    def _any():
        fun()

    return _any


FieldAllowedTypes = typing.Union[
    typing.List[AlgorithmProperty], typing.Type[BaseModel], typing.Type[AlgorithmDescribeBase]
]


class FormWidget(QWidget):
    value_changed = Signal()

    def __init__(
        self,
        fields: FieldAllowedTypes,
        start_values=None,
        dimension_num=1,
        settings: typing.Optional[BaseSettings] = None,
        parent=None,
    ):
        super().__init__(parent=parent)
        if start_values is None:
            start_values = {}
        self.widgets_dict: typing.Dict[str, QtAlgorithmProperty] = {}
        self.channels_chose: typing.List[typing.Union[ChannelComboBox, SubAlgorithmWidget]] = []
        layout = QFormLayout()
        layout.setContentsMargins(10, 0, 10, 0)
        self._model_class = None
        element_list = self._element_list(fields)
        for el in element_list:
            if isinstance(el, (QLabel, Hline)):
                layout.addRow(el)
                continue
            self._add_to_layout(layout, el, start_values, settings)
            if hasattr(el.get_field(), "change_channels_num"):
                self.channels_chose.append(el.get_field())
        self.setLayout(layout)
        self.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Minimum)

    def _element_list(self, fields: FieldAllowedTypes):
        if inspect.isclass(fields) and issubclass(fields, AlgorithmDescribeBase):
            if fields.__new_style__:
                self._model_class = fields.__argument_class__
                fields = base_model_to_algorithm_property(fields.__argument_class__)
            else:
                fields = fields.get_fields()
        elif not isinstance(fields, typing.Iterable):
            self._model_class = fields
            fields = base_model_to_algorithm_property(fields)
        return self._element_list_map(fields)

    @staticmethod
    def _element_list_map(fields):
        return map(QtAlgorithmProperty.from_algorithm_property, fields)

    def _add_to_layout(
        self, layout, ap: QtAlgorithmProperty, start_values: typing.MutableMapping, settings, add_to_widget_dict=True
    ):
        label = QLabel(ap.user_name)
        if ap.help_text:
            label.setToolTip(ap.help_text)
        if add_to_widget_dict:
            self.widgets_dict[ap.name] = ap
        ap.change_fun.connect(_any_arguments(self.value_changed.emit))
        if isinstance(ap.get_field(), SubAlgorithmWidget):
            w = typing.cast(SubAlgorithmWidget, ap.get_field())
            layout.addRow(label, w.choose)
            layout.addRow(ap.get_field())
            if ap.name in start_values:
                w.set_starting(start_values[ap.name])
            ap.change_fun.connect(_any_arguments(self.value_changed.emit))
            return
        if isinstance(ap.get_field(), FieldsList):
            layout.addRow(label)
            for el in typing.cast(FieldsList, ap.get_field()).field_list:
                self._add_to_layout(layout, el, start_values.get(ap.name, {}), settings, add_to_widget_dict=False)
            return
        if isinstance(ap.get_field(), Widget):
            widget = typing.cast(Widget, ap.get_field()).native
        else:
            widget = ap.get_field()
        if ap.is_multiline():
            layout.addRow(label)
            layout.addRow(widget)
        else:
            layout.addRow(label, widget)
        # noinspection PyUnresolvedReferences
        if inspect.isclass(ap.value_type) and issubclass(ap.value_type, ROIExtractionProfile):
            # noinspection PyTypeChecker
            ap.get_field().set_settings(settings)
        if ap.name in start_values:
            with suppress(KeyError, ValueError, TypeError):
                ap.set_value(start_values[ap.name])

    def has_elements(self):
        return len(self.widgets_dict) > 0

    def get_values(self):
        res = {name: el.get_value() for name, el in self.widgets_dict.items()}
        return self._model_class(**res) if self._model_class is not None else res

    def recursive_get_values(self):
        return {name: el.recursive_get_values() for name, el in self.widgets_dict.items()}

    def set_values(self, values: typing.Union[dict, BaseModel]):
        if isinstance(values, BaseModel):
            values = dict(values)
        for name, value in values.items():
            if name in self.widgets_dict:
                self.widgets_dict[name].set_value(value)

    def image_changed(self, image: Image):
        if not image:
            return
        for channel_widget in self.channels_chose:
            if isinstance(channel_widget, ChannelComboBox):
                channel_widget.change_channels_num(image.channels)
            else:
                channel_widget.change_channels_num(image)


class SubAlgorithmWidget(QWidget):
    values_changed = Signal()

    def __init__(self, algorithm_property: AlgorithmProperty):
        super().__init__()
        if not isinstance(algorithm_property.possible_values, typing.MutableMapping):
            raise ValueError(
                f"algorithm_property.possible_values should be dict. It is {type(algorithm_property.possible_values)}"
            )
        if not isinstance(algorithm_property.default_value, str):
            raise ValueError(
                f"algorithm_property.default_value should be str. It is {type(algorithm_property.default_value)}"
            )
        self.starting_values = {}
        self.property = algorithm_property
        self.widgets_dict: typing.Dict[str, FormWidget] = {}
        # TODO protect for recursion
        widget = self._get_form_widget(algorithm_property)
        widget.value_changed.connect(self.values_changed)

        self.widgets_dict[algorithm_property.default_value] = widget
        self.choose = QComboBox(self)
        self.choose.addItems(list(algorithm_property.possible_values.keys()))
        self.setContentsMargins(0, 0, 0, 0)

        self.choose.setCurrentText(algorithm_property.default_value)

        self.choose.currentTextChanged.connect(self.algorithm_choose)
        layout = QVBoxLayout()
        layout.setContentsMargins(4, 4, 4, 4)
        layout.addWidget(widget)
        if not widget.has_elements():
            widget.hide()
            self.hide()
        tmp_widget = QWidget(self)
        layout.addWidget(tmp_widget)
        self.tmp_widget = tmp_widget
        self.setLayout(layout)

    @staticmethod
    def _get_form_widget(algorithm_property, start_values=None):
        if isinstance(algorithm_property, AlgorithmProperty):
            calc_class = algorithm_property.possible_values[algorithm_property.default_value]
        else:
            calc_class = algorithm_property
        widget = FormWidget(calc_class, start_values=start_values)
        widget.layout().setContentsMargins(0, 0, 0, 0)
        return widget

    def set_starting(self, starting_values):
        self.starting_values = starting_values

    def set_values(self, val: typing.Mapping):
        if isinstance(val, BaseModel):
            val = dict(val)
        if not isinstance(val, typing.Mapping):
            return
        self.choose.setCurrentText(val["name"])
        if val["name"] not in self.widgets_dict:
            self.algorithm_choose(val["name"])
        if val["name"] in self.widgets_dict:
            self.widgets_dict[val["name"]].set_values(val["values"])

    def recursive_get_values(self):
        return {name: el.recursive_get_values() for name, el in self.widgets_dict.items()}

    def get_values(self) -> typing.Dict[str, typing.Any]:
        name = self.choose.currentText()
        values = self.widgets_dict[name].get_values()
        return {"name": name, "values": values}

    def change_channels_num(self, image: Image):
        for i in range(self.layout().count()):
            el = self.layout().itemAt(i)
            if el.widget() and isinstance(el.widget(), FormWidget):
                typing.cast(FormWidget, el.widget()).image_changed(image)

    def algorithm_choose(self, name):
        if name not in self.widgets_dict:
            if name not in self.property.possible_values:
                return
            start_dict = self.starting_values.get(name, {})
            self.widgets_dict[name] = self._get_form_widget(self.property.possible_values[name], start_dict)

            self.layout().addWidget(self.widgets_dict[name])
            self.widgets_dict[name].value_changed.connect(self.values_changed)
        widget = self.widgets_dict[name]
        for i in range(self.layout().count()):
            lay_elem = self.layout().itemAt(i)
            if lay_elem.widget():
                lay_elem.widget().setVisible(False)
        if widget.has_elements() and self.parent() is not None:
            self.setVisible(True)
            widget.setVisible(True)
        else:
            self.setVisible(False)
        self.values_changed.emit()

    def showEvent(self, _event):
        # workaround for changing size
        self.tmp_widget.hide()

    def paintEvent(self, event: QPaintEvent):
        name = self.choose.currentText()
        if self.widgets_dict[name].has_elements() and event.rect().top() == 0 and event.rect().left() == 0:
            painter = QPainter(self)
            painter.drawRect(self.rect() - QMargins(1, -1, 1, 1))


class BaseAlgorithmSettingsWidget(QScrollArea):
    values_changed = Signal()
    algorithm_thread: SegmentationThread

    def __init__(self, settings: BaseSettings, algorithm: typing.Type[ROIExtractionAlgorithm], parent=None):
        """
        For algorithm which works on one channel
        """
        super().__init__(parent=parent)
        self.settings = settings
        self.widget_list = []
        self.algorithm = algorithm
        main_layout = QVBoxLayout()
        self.info_label = QLabel()
        self.info_label.setHidden(True)
        # FIXME verify inflo_label usage
        main_layout.addWidget(self.info_label)
        start_values = settings.get_algorithm(f"algorithm_widget_state.{self.name}", {})
        self.form_widget = self._form_widget(algorithm, start_values=start_values)
        self.form_widget.value_changed.connect(self.values_changed.emit)
        self._widget = QWidget(self)
        self._widget.setLayout(QVBoxLayout())
        self._widget.layout().setContentsMargins(0, 0, 0, 0)
        self._widget.layout().addWidget(self.form_widget)
        self.setWidget(self._widget)
        value_dict = self.settings.get_algorithm(f"algorithms.{self.name}", {})
        self.set_values(value_dict)
        self.algorithm_thread = SegmentationThread(algorithm())
        self.algorithm_thread.info_signal.connect(self.show_info)
        self.algorithm_thread.exception_occurred.connect(self.exception_occurred)
        self.setWidgetResizable(True)

    @property
    def name(self):
        return self.algorithm.get_name()

    def _form_widget(self, algorithm, start_values) -> FormWidget:
        return FormWidget(
            algorithm,
            start_values=start_values,
            parent=self,
        )

    @staticmethod
    def exception_occurred(exc: Exception):
        if isinstance(exc, SegmentationLimitException):
            mess = QMessageBox()
            mess.setIcon(QMessageBox.Critical)
            mess.setText("During segmentation process algorithm meet limitations:\n" + "\n".join(exc.args))
            mess.setWindowTitle("Segmentation limitations")
            mess.exec_()
            return
        if isinstance(exc, RuntimeError) and exc.args[0].startswith(
            "Exception thrown in SimpleITK KittlerIllingworthThreshold"
        ):
            mess = QMessageBox()
            mess.setIcon(QMessageBox.Critical)
            mess.setText("Fail to apply Kittler Illingworth to current data\n" + exc.args[0].split("\n")[1])
            mess.setWindowTitle("Segmentation limitations")
            mess.exec_()
            return
        dial = ErrorDialog(exc, "Error during segmentation", f"{QApplication.instance().applicationName()}")
        dial.exec_()

    def show_info(self, text):
        self.info_label.setText(text)
        self.info_label.setVisible(bool(text))

    def image_changed(self, image: Image):
        self.form_widget.image_changed(image)
        self.algorithm_thread.set_image(image)

    def set_mask(self, mask):
        self.algorithm_thread.set_mask(mask)

    def mask(self) -> typing.Optional[np.ndarray]:
        return self.algorithm_thread.algorithm.mask

    def set_values(self, values_dict):
        self.form_widget.set_values(values_dict)

    def get_values(self):
        return self.form_widget.get_values()

    def execute(self, exclude_mask=None):
        values = self.get_values()
        self.settings.set_algorithm(f"algorithms.{self.name}", deepcopy(values))
        if isinstance(values, dict):
            self.algorithm_thread.set_parameters(**values)
        else:
            self.algorithm_thread.set_parameters(values)
        self.algorithm_thread.start()

    def hideEvent(self, a0: QHideEvent):
        self.algorithm_thread.clean()

    def resizeEvent(self, event: QResizeEvent):
        if self.height() < self.form_widget.height():
            self.setMinimumWidth(self.form_widget.width() + 20)
        else:
            self.setMinimumWidth(self.form_widget.width() + 10)
        super().resizeEvent(event)

    def recursive_get_values(self):
        return self.form_widget.recursive_get_values()


class InteractiveAlgorithmSettingsWidget(BaseAlgorithmSettingsWidget):
    algorithm_thread: SegmentationThread

    def __init__(
        self, settings, algorithm: typing.Type[ROIExtractionAlgorithm], selector: typing.List[QWidget], parent=None
    ):
        super().__init__(settings, algorithm, parent=parent)
        self.selector = selector[:]
        self.algorithm_thread.finished.connect(self.enable_selector)
        self.algorithm_thread.started.connect(self.disable_selector)
        # noinspection PyUnresolvedReferences
        if hasattr(settings, "mask_changed"):
            settings.mask_changed.connect(self.change_mask)

    def change_mask(self):
        if not self.isVisible():
            return
        self.algorithm_thread.algorithm.set_mask(self.settings.mask)

    def disable_selector(self):
        for el in self.selector:
            el.setDisabled(True)

    def enable_selector(self):
        for el in self.selector:
            el.setEnabled(True)

    def get_segmentation_profile(self) -> ROIExtractionProfile:
        return ROIExtractionProfile(name="", algorithm=self.algorithm.get_name(), values=self.get_values())


class AlgorithmChooseBase(QWidget):
    finished = Signal()
    started = Signal()
    result = Signal(ROIExtractionResult)
    value_changed = Signal()
    progress_signal = Signal(str, int)
    algorithm_changed = Signal(str)

    algorithm_dict: typing.Dict[str, InteractiveAlgorithmSettingsWidget]

    def __init__(self, settings: BaseSettings, algorithms: typing.Type[AlgorithmSelection], parent=None):
        super().__init__(parent=parent)
        self.settings = settings
        self.algorithms = algorithms
        settings.algorithm_changed.connect(self.updated_algorithm)
        self.stack_layout = QStackedLayout()
        self.algorithm_choose = QComboBox()
        self.algorithm_dict: typing.Dict[str, BaseAlgorithmSettingsWidget] = {}
        self.algorithm_choose.currentTextChanged.connect(self.change_algorithm)
        self.add_widgets_to_algorithm()

        self.setContentsMargins(0, 0, 0, 0)
        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.algorithm_choose)
        layout.addLayout(self.stack_layout)
        self.setLayout(layout)

    def _algorithm_widget(self, settings, val) -> InteractiveAlgorithmSettingsWidget:
        return InteractiveAlgorithmSettingsWidget(settings, val, [], parent=self)

    def add_widgets_to_algorithm(self):
        self.algorithm_choose.blockSignals(True)
        self.algorithm_choose.clear()
        for name, val in self.algorithms.__register__.items():
            self.algorithm_choose.addItem(name)
            widget = self._algorithm_widget(self.settings, val)
            self.algorithm_dict[name] = widget
            widget.algorithm_thread.execution_done.connect(self.result.emit)
            widget.algorithm_thread.finished.connect(self.finished.emit)
            widget.algorithm_thread.started.connect(self.started.emit)
            widget.algorithm_thread.progress_signal.connect(self.progress_signal.emit)
            widget.values_changed.connect(self.value_changed.emit)
            self.stack_layout.addWidget(widget)
        name = self.settings.get_algorithm("current_algorithm", "")
        self.algorithm_choose.blockSignals(False)
        if name:
            self.algorithm_choose.setCurrentText(name)

    def reload(self, algorithms=None):
        if algorithms is not None:
            self.algorithms = algorithms
        for _ in range(self.stack_layout.count()):
            widget = typing.cast(InteractiveAlgorithmSettingsWidget, self.stack_layout.takeAt(0).widget())
            widget.algorithm_thread.execution_done.disconnect()
            widget.algorithm_thread.finished.disconnect()
            widget.algorithm_thread.started.disconnect()
            widget.algorithm_thread.progress_signal.disconnect()
            widget.values_changed.disconnect()
            widget.deleteLater()
        self.algorithm_dict = {}
        self.add_widgets_to_algorithm()

    def updated_algorithm(self):
        self.change_algorithm(
            self.settings.last_executed_algorithm,
            self.settings.get_algorithm(f"algorithms.{self.settings.last_executed_algorithm}"),
        )

    def recursive_get_values(self):
        result = {key: widget.recursive_get_values() for key, widget in self.algorithm_dict.items()}

        self.settings.set_algorithm(
            "algorithm_widget_state",
            recursive_update(self.settings.get_algorithm("algorithm_widget_state", {}), result),
        )
        return result

    def change_algorithm(self, name, values: typing.Optional[dict] = None):
        self.settings.set_algorithm("current_algorithm", name)
        widget = typing.cast(InteractiveAlgorithmSettingsWidget, self.stack_layout.currentWidget())
        blocked = self.blockSignals(True)
        if name != widget.name:
            widget = self.algorithm_dict[name]
            self.stack_layout.setCurrentWidget(widget)
            widget.image_changed(self.settings.image)
            if hasattr(widget, "set_mask") and hasattr(self.settings, "mask"):
                widget.set_mask(self.settings.mask)
        elif values is None:
            self.blockSignals(blocked)
            return
        if values is not None:
            widget.set_values(values)
        self.algorithm_choose.setCurrentText(name)
        self.blockSignals(blocked)
        self.algorithm_changed.emit(name)

    def current_widget(self) -> InteractiveAlgorithmSettingsWidget:
        return typing.cast(InteractiveAlgorithmSettingsWidget, self.stack_layout.currentWidget())

    def current_parameters(self) -> ROIExtractionProfile:
        widget = self.current_widget()
        return ROIExtractionProfile(name="", algorithm=widget.name, values=widget.get_values())

    def get_info_text(self):
        return self.current_widget().algorithm_thread.get_info_text()


class AlgorithmChoose(AlgorithmChooseBase):
    def __init__(self, settings: BaseSettings, algorithms: typing.Type[AlgorithmSelection], parent=None):
        super().__init__(settings, algorithms, parent)
        self.settings.image_changed.connect(self.image_changed)

    def image_changed(self):
        current_widget = typing.cast(InteractiveAlgorithmSettingsWidget, self.stack_layout.currentWidget())
        prev_block = self.blockSignals(True)
        current_widget.image_changed(self.settings.image)
        if hasattr(self.settings, "mask") and hasattr(current_widget, "change_mask"):
            current_widget.change_mask()
        self.blockSignals(prev_block)
        self.value_changed.emit()


def _value_get(self: Widget):
    return self.value


def _value_set(self: Widget, value: typing.Any):
    if isinstance(self, ComboBox) and issubclass(self.annotation, Layer) and isinstance(value, str):
        for el in self.choices:
            if el.name == value:
                self.value = el
                return
    if isinstance(value, Channel):
        self.value = value.value
    self.value = value



import os

import numpy as np
from napari.utils import Colormap
from pytestqt.qtbot import QtBot
from qtpy.QtCore import QPoint, Qt
from qtpy.QtGui import QColor
from qtpy.QtWidgets import QFileDialog

import PartSegData
from PartSeg.common_backend.base_settings import BaseSettings, ColormapDict, ViewSettings
from PartSeg.common_gui.channel_control import ChannelProperty
from PartSeg.common_gui.colormap_creator import (
    ColormapCreator,
    ColormapEdit,
    ColormapList,
    PColormapCreator,
    PColormapCreatorMid,
    PColormapList,
    color_from_qcolor,
    qcolor_from_color,
)
from PartSeg.common_gui.napari_image_view import ImageView
from PartSegCore.color_image.base_colors import Color, starting_colors
from PartSegImage import TiffImageReader


def test_color_conversion():
    color1 = QColor(70, 52, 190)
    color2 = color_from_qcolor(color1)
    assert qcolor_from_color(color2) == color1
    assert color_from_qcolor(qcolor_from_color(color2)) == color2


class TestColormapEdit:
    def test_click(self, qtbot: QtBot, qapp):
        widget = ColormapEdit()
        widget.show()
        qtbot.addWidget(widget)
        width = widget.width() - 20
        with qtbot.waitSignal(widget.double_clicked):
            qtbot.mouseDClick(widget, Qt.LeftButton, pos=QPoint(30, widget.height() // 2))

        pos = 20 / width
        widget.add_color(pos, Color(red=0.274, green=0.02, blue=0.745))
        widget.add_color(pos / 2, Color(red=0.274, green=0.2, blue=0.745))
        assert len(widget.colormap.colors) == 4
        with qtbot.assertNotEmitted(widget.double_clicked):
            qtbot.mouseDClick(widget, Qt.LeftButton, pos=QPoint(30, widget.height() // 2))
        assert len(widget.colormap.colors) == 3
        qapp.processEvents()
        widget.hide()

    def test_distribute_evenly(self, qtbot: QtBot):
        widget = ColormapEdit()
        qtbot.addWidget(widget)
        widget.add_color(0.1, Color(red=0.49, green=0.9, blue=0.08))
        widget.add_color(0.23, Color(red=0.09, green=0.03, blue=0.788))
        widget.add_color(0.84, Color(red=0.874, green=0, blue=0.2))
        assert len(widget.colormap.colors) == 5
        widget.distribute_evenly()
        assert len(widget.colormap.colors) == 3
        assert widget.colormap.controls[0] == 0
        assert widget.colormap.controls[1] == 0.5
        assert widget.colormap.controls[2] == 1
        widget.clear()
        assert len(widget.colormap.colors) == 1

    def test_move(self, qtbot):
        widget = ColormapEdit()
        qtbot.addWidget(widget)
        width = widget.width() - 20
        pos = 20 / width
        widget.add_color(pos, Color(red=0.49, green=0.9, blue=0.08))
        assert np.isclose(widget.colormap.controls[1], pos)
        qtbot.mousePress(widget, Qt.MouseButton.LeftButton, pos=QPoint(30, widget.height() // 2))
        pos2 = 150 / width
        qtbot.mouseMove(widget, QPoint(70, widget.height() // 2))
        qtbot.mouseMove(widget, QPoint(160, widget.height() // 2))
        qtbot.mouseRelease(widget, Qt.MouseButton.LeftButton, pos=QPoint(160, widget.height() // 2))
        assert np.isclose(widget.colormap.controls[1], pos2)


class TestColormapCreator:
    def test_add_color(self, qtbot: QtBot):
        widget = ColormapCreator()
        colormap_edit = widget.show_colormap
        color1 = QColor(10, 40, 12)
        color2 = QColor(100, 4, 220)
        widget.color_picker.setCurrentColor(color1)
        with qtbot.waitSignal(colormap_edit.double_clicked):
            qtbot.mouseDClick(colormap_edit, Qt.LeftButton, pos=QPoint(30, widget.height() // 2))
        assert len(widget.current_colormap().colors) == 3
        assert np.allclose(
            widget.current_colormap().colors[0], Color(red=10 / 255, green=40 / 255, blue=12 / 255).as_tuple()
        )
        assert np.isclose(widget.current_colormap().controls[1], 20 / (colormap_edit.width() - 20))
        widget.color_picker.setCurrentColor(color2)
        with qtbot.waitSignal(colormap_edit.double_clicked):
            qtbot.mouseDClick(colormap_edit, Qt.LeftButton, pos=QPoint(80, widget.height() // 2))
        assert len(widget.current_colormap().colors) == 4
        assert np.allclose(
            widget.current_colormap().colors[1], Color(red=10 / 255, green=40 / 255, blue=12 / 255).as_tuple()
        )
        assert np.allclose(
            widget.current_colormap().colors[2], Color(red=100 / 255, green=4 / 255, blue=220 / 255).as_tuple()
        )
        assert np.isclose(widget.current_colormap().controls[1], 20 / (colormap_edit.width() - 20))
        assert np.isclose(widget.current_colormap().controls[2], 70 / (colormap_edit.width() - 20))

    def test_save(self, qtbot):
        widget = ColormapCreator()
        qtbot.addWidget(widget)
        color1 = QColor(10, 40, 12)
        color2 = QColor(100, 4, 220)
        widget.color_picker.setCurrentColor(color1)
        widget.add_color(0.1)
        widget.color_picker.setCurrentColor(color2)
        widget.add_color(0.8)

        def check_res(colormap):
            return (
                len(colormap.colors) == 4
                and np.allclose(colormap.colors[1], color_from_qcolor(color1).as_tuple())
                and np.allclose(colormap.colors[2], color_from_qcolor(color2).as_tuple())
                and np.isclose(colormap.controls[1], 0.1)
                and np.isclose(colormap.controls[2], 0.8)
            )

        with qtbot.wait_signal(widget.colormap_selected, check_params_cb=check_res):
            widget.save_btn.click()

    def test_distribute(self, qtbot):
        widget = ColormapCreator()
        qtbot.addWidget(widget)
        color1 = QColor(10, 40, 12)
        color2 = QColor(100, 4, 220)
        color3 = QColor(100, 24, 220)
        color4 = QColor(100, 134, 22)
        color5 = QColor(100, 134, 122)
        widget.color_picker.setCurrentColor(color1)
        widget.add_color(0.1)
        widget.color_picker.setCurrentColor(color2)
        widget.add_color(0.3)
        widget.color_picker.setCurrentColor(color3)
        widget.add_color(0.6)
        widget.color_picker.setCurrentColor(color4)
        widget.add_color(0.9)
        widget.color_picker.setCurrentColor(color5)
        widget.add_color(0.95)
        li = widget.current_colormap()
        assert np.allclose(li.controls, [0, 0.1, 0.3, 0.6, 0.9, 0.95, 1])
        for el, col in zip(
            li.colors,
            [
                color_from_qcolor(color1),
                color_from_qcolor(color1),
                color_from_qcolor(color2),
                color_from_qcolor(color3),
                color_from_qcolor(color4),
                color_from_qcolor(color5),
                color_from_qcolor(color5),
            ],
        ):
            assert np.allclose(el, col.as_tuple())

        widget.distribute_btn.click()
        li = widget.current_colormap()
        assert np.allclose(li.controls, [0, 0.25, 0.5, 0.75, 1])
        for el, col in zip(
            li.colors,
            [
                color_from_qcolor(color1),
                color_from_qcolor(color2),
                color_from_qcolor(color3),
                color_from_qcolor(color4),
                color_from_qcolor(color5),
            ],
        ):
            assert np.allclose(el, col.as_tuple())

    def test_colormap_creator_save(self, qtbot, tmp_path, monkeypatch, base_settings):
        widget = PColormapCreatorMid(base_settings)
        cmap = Colormap([[0, 0, 0], [0.5, 0.5, 0.5], [1, 1, 1]], controls=[0, 0.3, 1])
        widget.show_colormap.colormap = cmap
        qtbot.addWidget(widget)
        target_path = str(tmp_path / "test.json")

        def exec_(self_):
            self_.selectFile(target_path)
            self_.accept()
            return True

        def selected_name_filter(self_):
            return self_.nameFilters()[0]

        def selected_files(self_):
            return [target_path]

        monkeypatch.setattr(QFileDialog, "exec_", exec_)
        monkeypatch.setattr(QFileDialog, "selectedNameFilter", selected_name_filter)
        monkeypatch.setattr(QFileDialog, "selectedFiles", selected_files)

        widget._export_action()
        assert os.path.isfile(target_path)

        widget.clear_btn.click()
        assert widget.show_colormap.colormap != cmap

        widget._import_action()

        assert widget.show_colormap.colormap == cmap


class TestPColormapCreator:
    def test_save(self, qtbot, base_settings):
        widget = PColormapCreator(base_settings)
        qtbot.addWidget(widget)
        color1 = QColor(10, 40, 12)
        color2 = QColor(100, 4, 220)
        widget.color_picker.setCurrentColor(color1)
        widget.add_color(0.1)
        widget.color_picker.setCurrentColor(color2)
        widget.add_color(0.8)

        def check_res(colormap):
            return (
                len(colormap.colors) == 4
                and np.allclose(colormap.colors[0], color_from_qcolor(color1).as_tuple())
                and np.allclose(colormap.colors[1], color_from_qcolor(color1).as_tuple())
                and np.allclose(colormap.colors[2], color_from_qcolor(color2).as_tuple())
                and np.allclose(colormap.colors[3], color_from_qcolor(color2).as_tuple())
                and np.allclose(colormap.controls, [0, 0.1, 0.8, 1])
            )

        with qtbot.wait_signal(widget.colormap_selected, check_params_cb=check_res):
            widget.save_btn.click()

        cmap_dict = base_settings.get_from_profile("custom_colormap")
        assert len(cmap_dict) == 1
        assert check_res(next(iter(cmap_dict.values())))

    def test_custom_colors_save(self, qtbot, base_settings):
        widget = PColormapCreator(base_settings)
        qtbot.addWidget(widget)
        color1 = QColor(10, 40, 12)
        color2 = QColor(100, 4, 220)
        widget.color_picker.setCustomColor(2, color1)
        widget.color_picker.setCustomColor(7, color2)
        widget._save_custom_colors()
        custom_color_list = base_settings.get_from_profile("custom_colors")
        assert custom_color_list[2] == color_from_qcolor(color1)
        assert custom_color_list[7] == color_from_qcolor(color2)
        widget2 = PColormapCreator(base_settings)
        assert widget2.color_picker.customColor(2) == color1
        assert widget2.color_picker.customColor(7) == color2


class TestColormapList:
    @staticmethod
    def verify_visibility(name, state):
        def _check(_name, _state):
            return name == _name and state == _state

        return _check

    def test_base(self, qtbot):
        dkt = ColormapDict({})
        widget = ColormapList(dkt, starting_colors)
        widget.refresh()
        qtbot.addWidget(widget)
        with qtbot.waitSignal(
            widget.visibility_colormap_change, check_params_cb=self.verify_visibility(starting_colors[0], False)
        ):
            widget.set_state(starting_colors[0], False)
        assert len(widget.get_selected()) == len(starting_colors) - 1
        with qtbot.waitSignal(
            widget.visibility_colormap_change, check_params_cb=self.verify_visibility(starting_colors[0], True)
        ):
            widget.set_state(starting_colors[0], True)
        assert len(widget.get_selected()) == len(starting_colors)
        name = f"{starting_colors[0]}_reversed"
        with qtbot.waitSignal(widget.visibility_colormap_change, check_params_cb=self.verify_visibility(name, True)):
            widget.set_state(name, True)
        assert len(widget.get_selected()) == len(starting_colors) + 1

    def test_edit_button(self, qtbot):
        dkt = ColormapDict({})
        color_list = ColormapList(dkt, starting_colors)
        color_list.refresh()
        qtbot.addWidget(color_list)
        color_edit = ColormapCreator()
        color_list.edit_signal.connect(color_edit.set_colormap)
        qtbot.addWidget(color_edit)
        cmap = dkt["red"][0]
        assert np.any(cmap.colors != color_edit.current_colormap().colors)
        color_list.get_colormap_widget("red").edit_btn.click()
        assert np.all(cmap.colors == color_edit.current_colormap().colors)
        assert color_list._widget_dict["magma"].edit_btn.isEnabled() is False

    def test_settings_integration(self, qtbot):
        settings = ViewSettings()
        color_list = PColormapList(settings, [])
        color_list.refresh()
        qtbot.addWidget(color_list)
        selected = settings.chosen_colormap[:]
        color_list.set_state("red", False)
        selected2 = settings.chosen_colormap[:]
        assert len(selected2) + 1 == len(selected)
        assert "red" not in selected2
        assert "red" in selected

    def test_image_view_integration(self, qtbot, tmp_path):
        settings = BaseSettings(tmp_path)
        channel_property = ChannelProperty(settings, "test")
        image_view = ImageView(settings, channel_property, "test")
        qtbot.addWidget(channel_property)
        qtbot.addWidget(image_view)
        color_list = PColormapList(settings, ["test"])
        qtbot.addWidget(color_list)
        image = TiffImageReader.read_image(PartSegData.segmentation_analysis_default_image)
        with qtbot.wait_signal(image_view.image_added, timeout=10**6):
            settings.image = image
        color_list.refresh()
        assert image_view.channel_control.channels_count == image.channels
        assert len(color_list.blocked()) == image.channels
        block_count = 0
        for el in settings.colormap_dict:
            widget = color_list.get_colormap_widget(el)
            assert widget.is_checked or el not in starting_colors
            if not widget.checked.isEnabled():
                block_count += 1
        assert block_count == image.channels
        image_view.channel_control.change_selected_color(0, "gray")
        assert len(color_list.blocked()) == image.channels
        assert "gray" in color_list.blocked()
        color_list.refresh()
        assert color_list.get_colormap_widget("gray").checked.isEnabled() is False
        # this lines test if after refresh of widget checkbox stays checkable
        block_count = 0
        for el in settings.colormap_dict:
            widget = color_list.get_colormap_widget(el)
            assert widget.is_checked or el not in starting_colors
            if not widget.checked.isEnabled():
                block_count += 1
        assert block_count == image.channels



from contextlib import suppress

from qtpy.QtCore import QByteArray, QEvent, Qt
from qtpy.QtGui import QCloseEvent, QKeyEvent
from qtpy.QtWidgets import (
    QApplication,
    QCheckBox,
    QHBoxLayout,
    QLabel,
    QMessageBox,
    QPushButton,
    QTableWidget,
    QTableWidgetItem,
    QVBoxLayout,
    QWidget,
)
from superqt import QEnumComboBox

from PartSeg._roi_mask.stack_settings import StackSettings
from PartSeg.common_gui.universal_gui_part import ChannelComboBox
from PartSeg.common_gui.waiting_dialog import ExecuteFunctionDialog
from PartSegCore import Units
from PartSegCore.algorithm_describe_base import base_model_to_algorithm_property
from PartSegCore.analysis.measurement_base import AreaType, Leaf, MeasurementEntry, PerComponent
from PartSegCore.analysis.measurement_calculation import MEASUREMENT_DICT, MeasurementProfile, MeasurementResult


class SimpleMeasurements(QWidget):
    def __init__(self, settings: StackSettings, parent=None):
        super().__init__(parent)
        self.settings = settings
        self.calculate_btn = QPushButton("Calculate")
        self.calculate_btn.clicked.connect(self.calculate)
        self.result_view = QTableWidget()
        self.channel_select = ChannelComboBox()
        self.units_select = QEnumComboBox(enum_class=Units)
        self.units_select.setCurrentEnum(self.settings.get("simple_measurements.units", Units.nm))
        self.units_select.currentIndexChanged.connect(self.change_units)
        self._shift = 2

        layout = QHBoxLayout()
        self.measurement_layout = QVBoxLayout()
        l1 = QHBoxLayout()
        l1.addWidget(QLabel("Units"))
        l1.addWidget(self.units_select)
        self.measurement_layout.addLayout(l1)
        l2 = QHBoxLayout()
        l2.addWidget(QLabel("Channel"))
        l2.addWidget(self.channel_select)
        self.measurement_layout.addLayout(l2)
        layout.addLayout(self.measurement_layout)
        result_layout = QVBoxLayout()
        result_layout.addWidget(self.result_view)
        result_layout.addWidget(self.calculate_btn)
        layout.addLayout(result_layout)
        self.setLayout(layout)
        self.setWindowTitle("Measurement")
        if self.window() == self:
            with suppress(KeyError):
                geometry = self.settings.get_from_profile("simple_measurement_window_geometry")
                self.restoreGeometry(QByteArray.fromHex(bytes(geometry, "ascii")))

    def closeEvent(self, event: QCloseEvent) -> None:
        """
        Save geometry if widget is used as standalone window.
        """
        if self.window() == self:
            self.settings.set_in_profile(
                "simple_measurement_window_geometry", self.saveGeometry().toHex().data().decode("ascii")
            )
        super().closeEvent(event)

    def calculate(self):
        if self.settings.roi is None:
            QMessageBox.warning(self, "No segmentation", "need segmentation to work")
            return
        to_calculate = []
        for i in range(self._shift, self.measurement_layout.count()):
            # noinspection PyTypeChecker
            chk: QCheckBox = self.measurement_layout.itemAt(i).widget()
            if chk.isChecked():
                leaf: Leaf = MEASUREMENT_DICT[chk.text()].get_starting_leaf()
                to_calculate.append(leaf.replace_(per_component=PerComponent.Yes, area=AreaType.ROI))
        if not to_calculate:
            QMessageBox.warning(self, "No measurement", "Select at least one measurement")
            return

        profile = MeasurementProfile(
            name="", chosen_fields=[MeasurementEntry(name=x.name, calculation_tree=x) for x in to_calculate]
        )

        dial = ExecuteFunctionDialog(
            profile.calculate,
            kwargs={
                "image": self.settings.image,
                "channel_num": self.channel_select.get_value(),
                "roi": self.settings.roi_info,
                "result_units": self.units_select.currentEnum(),
            },
        )
        dial.exec_()
        result: MeasurementResult = dial.get_result()
        values = result.get_separated()
        labels = result.get_labels()
        self.result_view.clear()
        self.result_view.setColumnCount(len(values) + 1)
        self.result_view.setRowCount(len(labels))
        for i, val in enumerate(labels):
            self.result_view.setItem(i, 0, QTableWidgetItem(val))
        for j, values_list in enumerate(values):
            for i, val in enumerate(values_list):
                self.result_view.setItem(i, j + 1, QTableWidgetItem(str(val)))

    def _clean_measurements(self):
        selected = set()
        for _ in range(self.measurement_layout.count() - self._shift):
            # noinspection PyTypeChecker
            chk: QCheckBox = self.measurement_layout.takeAt(self._shift).widget()
            if chk.isChecked():
                selected.add(chk.text())
            chk.deleteLater()
        return selected

    def refresh_measurements(self):
        selected = self._clean_measurements()
        for val in MEASUREMENT_DICT.values():
            area = val.get_starting_leaf().area
            pc = val.get_starting_leaf().per_component
            if (
                base_model_to_algorithm_property(val.__argument_class__)
                if val.__new_style__
                else val.get_fields()
                or (area is not None and area != AreaType.ROI)
                or (pc is not None and pc != PerComponent.Yes)
            ):
                continue
            text = val.get_name()
            chk = QCheckBox(text)
            if text in selected:
                chk.setChecked(True)
            self.measurement_layout.addWidget(chk)

    def keyPressEvent(self, e: QKeyEvent):
        if not e.modifiers() & Qt.KeyboardModifier.ControlModifier:
            return
        selected = self.result_view.selectedRanges()

        if e.key() == Qt.Key.Key_C:  # copy
            s = ""

            for r in range(selected[0].topRow(), selected[0].bottomRow() + 1):
                for c in range(selected[0].leftColumn(), selected[0].rightColumn() + 1):
                    try:
                        s += str(self.result_view.item(r, c).text()) + "\t"
                    except AttributeError:  # noqa: PERF203
                        s += "\t"
                s = s[:-1] + "\n"  # eliminate last '\t'
            QApplication.clipboard().setText(s)

    def event(self, event: QEvent) -> bool:
        if event.type() == QEvent.Type.WindowActivate:
            if self.settings.image is not None:
                self.channel_select.change_channels_num(self.settings.image.channels)
            self.refresh_measurements()

        return super().event(event)

    def change_units(self):
        self.settings.set("simple_measurements.units", self.units_select.currentEnum())



from typing import Callable, Dict, Optional

from qtpy.QtCore import QEvent
from qtpy.QtWidgets import QGridLayout, QLabel, QListWidget, QPlainTextEdit, QPushButton, QWidget

from PartSeg._roi_mask.stack_settings import StackSettings
from PartSegCore.algorithm_describe_base import ROIExtractionProfile
from PartSegCore.mask.algorithm_description import MaskAlgorithmSelection


class SegmentationInfoDialog(QWidget):
    def __init__(self, settings: StackSettings, set_parameters: Callable[[str, dict], None], additional_text=None):
        """

        :param settings:
        :param set_parameters: Function which set parameters of chosen in dialog.
        :param additional_text: Additional text on top of Window.
        """
        super().__init__()
        self.settings = settings
        self.parameters_dict = None
        self.set_parameters = set_parameters
        self.components = QListWidget()
        self.components.currentItemChanged.connect(self.change_component_info)
        self.description = QPlainTextEdit()
        self.description.setReadOnly(True)
        self.close_btn = QPushButton("Close")
        self.close_btn.clicked.connect(self.close)
        self.set_parameters_btn = QPushButton("Reuse parameters")
        self.set_parameters_btn.clicked.connect(self.set_parameter_action)
        self.additional_text_label = QLabel(additional_text)
        self.additional_text_label.setVisible(bool(additional_text))
        self.additional_text_label.setWordWrap(True)
        layout = QGridLayout()
        layout.addWidget(self.additional_text_label, 0, 0, 1, 2)
        if not additional_text:
            self.additional_text_label.setVisible(False)

        layout.addWidget(QLabel("Components:"), 1, 0)
        layout.addWidget(QLabel("segmentation parameters:"), 1, 1)
        layout.addWidget(self.components, 2, 0)
        layout.addWidget(self.description, 2, 1)
        layout.addWidget(self.close_btn, 3, 0)
        layout.addWidget(self.set_parameters_btn, 3, 1)
        self.setLayout(layout)
        self.setWindowTitle("Parameters preview")

    def set_parameters_dict(self, val: Optional[Dict[int, ROIExtractionProfile]]):
        self.parameters_dict = val

    def set_additional_text(self, text):
        self.additional_text_label.setText(text)
        self.additional_text_label.setVisible(bool(text))

    @property
    def get_parameters(self):
        return self.parameters_dict or self.settings.components_parameters_dict

    def change_component_info(self):
        if self.components.currentItem() is None:
            return
        text = self.components.currentItem().text()
        parameters = self.get_parameters[int(text)]
        if parameters is None:
            self.description.setPlainText("None")
        else:
            self.description.setPlainText(f"Component {text}\n{parameters.pretty_print(MaskAlgorithmSelection)}")

    def set_parameter_action(self):
        if self.components.currentItem() is None:
            return
        text = self.components.currentItem().text()
        parameters = self.get_parameters[int(text)]
        self.set_parameters(parameters.algorithm, parameters.values)

    def event(self, event: QEvent):
        if event.type() == QEvent.Type.WindowActivate:
            index = self.components.currentRow()
            self.components.clear()
            self.components.addItems(list(map(str, self.get_parameters.keys())))
            self.components.setCurrentRow(index)
        return super().event(event)
